from unittest import IsolatedAsyncioTestCase

import orjson as json

from panther import Panther
from panther.app import API, GenericAPI
from panther.configs import config
from panther.request import Request
from panther.response import Response
from panther.test import APIClient


@API()
async def request_path(request: Request):
    return request.path


@API()
async def request_client(request: Request):
    return request.client.__dict__


@API()
async def request_query_params(request: Request):
    return request.query_params


@API()
async def request_data(request: Request):
    return request.data


@API()
async def request_path_variables(name: str, age: int, is_alive: bool):
    return {'name': name, 'age': age, 'is_alive': is_alive}


@API()
async def request_header(request: Request):
    return request.headers.__dict__


@API()
async def request_header_str(request: Request):
    return request.headers.__str__(), request.headers.__repr__()


@API()
async def request_header_contains(request: Request, header: str):
    return header in request.headers


@API()
async def request_header_by_attr(request: Request):
    return request.headers.authorization


@API()
async def request_header_by_item(request: Request):
    return request.headers['Authorization'], request.headers['authorization']


# # # Methods
class AllMethods(GenericAPI):
    def get(self, *args, **kwargs):
        return Response()

    def post(self, *args, **kwargs):
        return Response()

    def put(self, *args, **kwargs):
        return Response()

    def patch(self, *args, **kwargs):
        return Response()

    def delete(self, *args, **kwargs):
        return Response()


@API()
async def all_methods():
    return Response()


class GetMethod(GenericAPI):
    def get(self, *args, **kwargs):
        return Response()


@API(methods=['GET'])
async def get_method():
    return Response()


class PostMethod(GenericAPI):
    def post(self, *args, **kwargs):
        return Response()


@API(methods=['POST'])
async def post_method():
    return Response()


class PutMethod(GenericAPI):
    def put(self, *args, **kwargs):
        return Response()


@API(methods=['PUT'])
async def put_method():
    return Response()


class PatchMethod(GenericAPI):
    def patch(self, *args, **kwargs):
        return Response()


@API(methods=['PATCH'])
async def patch_method():
    return Response()


class DeleteMethod(GenericAPI):
    def delete(self, *args, **kwargs):
        return Response()


@API(methods=['DELETE'])
async def delete_method():
    return Response()


class GetPostPatchMethods(GenericAPI):
    def get(self, *args, **kwargs):
        return Response()

    def post(self, *args, **kwargs):
        return Response()

    def patch(self, *args, **kwargs):
        return Response()


@API(methods=['GET', 'POST', 'PATCH'])
async def get_post_patch_methods():
    return Response()


urls = {
    'path': request_path,
    'client': request_client,
    'query-params': request_query_params,
    'data': request_data,
    'path/<name>/variable/<age>/<is_alive>/': request_path_variables,
    'header': request_header,
    'header-str': request_header_str,
    'header-attr': request_header_by_attr,
    'header-item': request_header_by_item,
    'header-contains/<header>/': request_header_contains,
    'all-func': all_methods,
    'all-class': AllMethods,
    'get-func': get_method,
    'get-class': GetMethod,
    'post-func': post_method,
    'post-class': PostMethod,
    'put-func': put_method,
    'put-class': PutMethod,
    'patch-func': patch_method,
    'patch-class': PatchMethod,
    'delete-func': delete_method,
    'delete-class': DeleteMethod,
    'get-post-patch-func': get_post_patch_methods,
    'get-post-patch-class': GetPostPatchMethods,
}


class TestRequest(IsolatedAsyncioTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        app = Panther(__name__, configs=__name__, urls=urls)
        cls.client = APIClient(app=app)

    @classmethod
    def tearDownClass(cls) -> None:
        config.refresh()

    async def test_path(self):
        res = await self.client.get('path/')
        assert res.status_code == 200
        assert res.data == '/path/'

    async def test_client(self):
        res = await self.client.get('client/')
        assert res.status_code == 200
        assert res.data == {'ip': '127.0.0.1', 'port': 8585}

    async def test_query_params(self):
        res = await self.client.get(
            'query-params/',
            query_params={'my': 'name', 'is': 'ali', 'how': 'are'},
        )
        assert res.status_code == 200
        assert res.data == {'my': 'name', 'is': 'ali', 'how': 'are'}

    async def test_data(self):
        payload = {'detail': 'ok'}
        res = await self.client.post('data/', payload=json.dumps(payload))
        assert res.status_code == 200
        assert res.data == payload

    async def test_path_variables(self):
        res = await self.client.post('path/Ali/variable/27/true/')
        expected_response = {'name': 'Ali', 'age': 27, 'is_alive': True}
        assert res.status_code == 200
        assert res.data == expected_response

    # # # Headers
    async def test_headers_none(self):
        res = await self.client.get('header')
        expected_headers = {}
        assert res.data == expected_headers

    async def test_headers_content_type(self):
        res = await self.client.post('header')
        expected_headers = {'content-type': 'application/json'}
        assert res.data == expected_headers

    async def test_headers_full_items(self):
        headers = {
            'User-Agent': 'PostmanRuntime/7.36.0',
            'Accept': '*/*',
            'Cache-Control': 'no-cache',
            'Host': '127.0.0.1:8000',
            'Accept-Encoding': 'gzip, deflate, br',
            'Connection': 'keep-alive',
            'Content-Length': 546,
        }
        res = await self.client.post('header', headers=headers)
        expected_headers = {
            'content-type': 'application/json',
            'User-Agent': 'PostmanRuntime/7.36.0',
            'Accept': '*/*',
            'Cache-Control': 'no-cache',
            'Host': '127.0.0.1:8000',
            'Accept-Encoding': 'gzip, deflate, br',
            'Connection': 'keep-alive',
            'Content-Length': '546',
        }
        assert res.data == expected_headers

    async def test_headers_unknown_items(self):
        headers = {
            'Header1': 'PostmanRuntime/7.36.0',
            'Header2': '*/*',
        }
        res = await self.client.post('header', headers=headers)
        expected_headers = {
            'content-type': 'application/json',
            'Header1': 'PostmanRuntime/7.36.0',
            'Header2': '*/*',
        }
        assert res.data == expected_headers

    async def test_headers_authorization_by_getattr(self):
        headers = {
            'Authorization': 'Token xxx',
        }
        res = await self.client.post('header-attr', headers=headers)
        assert res.data == 'Token xxx'

    async def test_headers_authorization_by_getitem(self):
        headers = {
            'Authorization': 'Token xxx',
        }
        res = await self.client.post('header-item', headers=headers)
        assert res.data == ['Token xxx', 'Token xxx']

    async def test_headers_str(self):
        res = await self.client.post('header-str')
        assert res.data == ['Headers(content-type=application/json)', 'Headers(content-type=application/json)']

    async def test_headers_contains(self):
        res = await self.client.post('header-contains/content-type/')
        assert res.data is True
        res = await self.client.post('header-contains/accept/')
        assert res.data is False

    # # # Methods
    async def test_method_all(self):
        res_func = await self.client.get('all-func/')
        res_class = await self.client.get('all-class/')
        assert res_func.status_code == 200
        assert res_class.status_code == 200

        res_func = await self.client.post('all-func/')
        res_class = await self.client.post('all-class/')
        assert res_func.status_code == 200
        assert res_class.status_code == 200

        res_func = await self.client.put('all-func/')
        res_class = await self.client.put('all-class/')
        assert res_func.status_code == 200
        assert res_class.status_code == 200

        res_func = await self.client.patch('all-func/')
        res_class = await self.client.patch('all-class/')
        assert res_func.status_code == 200
        assert res_class.status_code == 200

        res_func = await self.client.delete('all-func/')
        res_class = await self.client.delete('all-class/')
        assert res_func.status_code == 200
        assert res_class.status_code == 200

    async def test_method_get(self):
        res_func = await self.client.get('get-func/')
        res_class = await self.client.get('get-class/')
        assert res_func.status_code == 200
        assert res_class.status_code == 200

        res_func = await self.client.post('get-func/')
        res_class = await self.client.post('get-class/')
        assert res_func.status_code == 405
        assert res_class.status_code == 405

        res_func = await self.client.put('get-func/')
        res_class = await self.client.put('get-class/')
        assert res_func.status_code == 405
        assert res_class.status_code == 405

        res_func = await self.client.patch('get-func/')
        res_class = await self.client.patch('get-class/')
        assert res_func.status_code == 405
        assert res_class.status_code == 405

        res_func = await self.client.delete('get-func/')
        res_class = await self.client.delete('get-class/')
        assert res_func.status_code == 405
        assert res_class.status_code == 405

    async def test_method_post(self):
        res_func = await self.client.get('post-func/')
        res_class = await self.client.get('post-class/')
        assert res_func.status_code == 405
        assert res_class.status_code == 405

        res_func = await self.client.post('post-func/')
        res_class = await self.client.post('post-class/')
        assert res_func.status_code == 200
        assert res_class.status_code == 200

        res_func = await self.client.put('post-func/')
        res_class = await self.client.put('post-class/')
        assert res_func.status_code == 405
        assert res_class.status_code == 405

        res_func = await self.client.patch('post-func/')
        res_class = await self.client.patch('post-class/')
        assert res_func.status_code == 405
        assert res_class.status_code == 405

        res_func = await self.client.delete('post-func/')
        res_class = await self.client.delete('post-class/')
        assert res_func.status_code == 405
        assert res_class.status_code == 405

    async def test_method_put(self):
        res_func = await self.client.get('put-func/')
        res_class = await self.client.get('put-class/')
        assert res_func.status_code == 405
        assert res_class.status_code == 405

        res_func = await self.client.post('put-func/')
        res_class = await self.client.post('put-class/')
        assert res_func.status_code == 405
        assert res_class.status_code == 405

        res_func = await self.client.put('put-func/')
        res_class = await self.client.put('put-class/')
        assert res_func.status_code == 200
        assert res_class.status_code == 200

        res_func = await self.client.patch('put-func/')
        res_class = await self.client.patch('put-class/')
        assert res_func.status_code == 405
        assert res_class.status_code == 405

        res_func = await self.client.delete('put-func/')
        res_class = await self.client.delete('put-class/')
        assert res_func.status_code == 405
        assert res_class.status_code == 405

    async def test_method_patch(self):
        res_func = await self.client.get('patch-func/')
        res_class = await self.client.get('patch-class/')
        assert res_func.status_code == 405
        assert res_class.status_code == 405

        res_func = await self.client.post('patch-func/')
        res_class = await self.client.post('patch-class/')
        assert res_func.status_code == 405
        assert res_class.status_code == 405

        res_func = await self.client.put('patch-func/')
        res_class = await self.client.put('patch-class/')
        assert res_func.status_code == 405
        assert res_class.status_code == 405

        res_func = await self.client.patch('patch-func/')
        res_class = await self.client.patch('patch-class/')
        assert res_func.status_code == 200
        assert res_class.status_code == 200

        res_func = await self.client.delete('patch-func/')
        res_class = await self.client.delete('patch-class/')
        assert res_func.status_code == 405
        assert res_class.status_code == 405

    async def test_method_delete(self):
        res_func = await self.client.get('delete-func/')
        res_class = await self.client.get('delete-class/')
        assert res_func.status_code == 405
        assert res_class.status_code == 405

        res_func = await self.client.post('delete-func/')
        res_class = await self.client.post('delete-class/')
        assert res_func.status_code == 405
        assert res_class.status_code == 405

        res_func = await self.client.put('delete-func/')
        res_class = await self.client.put('delete-class/')
        assert res_func.status_code == 405
        assert res_class.status_code == 405

        res_func = await self.client.patch('delete-func/')
        res_class = await self.client.patch('delete-class/')
        assert res_func.status_code == 405
        assert res_class.status_code == 405

        res_func = await self.client.delete('delete-func/')
        res_class = await self.client.delete('delete-class/')
        assert res_func.status_code == 200
        assert res_class.status_code == 200

    async def test_method_get_post_patch(self):
        res_func = await self.client.get('get-post-patch-func/')
        res_class = await self.client.get('get-post-patch-class/')
        assert res_func.status_code == 200
        assert res_class.status_code == 200

        res_func = await self.client.post('get-post-patch-func/')
        res_class = await self.client.post('get-post-patch-class/')
        assert res_func.status_code == 200
        assert res_class.status_code == 200

        res_func = await self.client.put('get-post-patch-func/')
        res_class = await self.client.put('get-post-patch-class/')
        assert res_func.status_code == 405
        assert res_class.status_code == 405

        res_func = await self.client.patch('get-post-patch-func/')
        res_class = await self.client.patch('get-post-patch-class/')
        assert res_func.status_code == 200
        assert res_class.status_code == 200

        res_func = await self.client.delete('get-post-patch-func/')
        res_class = await self.client.delete('get-post-patch-class/')
        assert res_func.status_code == 405
        assert res_class.status_code == 405



import logging
from time import perf_counter

from panther.base_websocket import Websocket
from panther.middlewares import HTTPMiddleware, WebsocketMiddleware
from panther.request import Request

logger = logging.getLogger('monitoring')


class MonitoringMiddleware(HTTPMiddleware):
    """
    Create Log Message Like Below:
    datetime | method | path | ip:port | response_time(seconds) | status
    """

    async def __call__(self, request: Request):
        start_time = perf_counter()
        method = request.scope['method']

        response = await self.dispatch(request=request)

        response_time = perf_counter() - start_time  # Seconds
        logger.info(f'{method} | {request.path} | {request.client} | {response_time} | {response.status_code}')
        return response


class WebsocketMonitoringMiddleware(WebsocketMiddleware):
    """
    Create Log Message Like Below:
    datetime | WS | path | ip:port | connection_time(seconds) | status
    """

    ConnectedConnectionTime = ' - '

    async def __call__(self, connection: Websocket):
        start_time = perf_counter()

        logger.info(f'WS | {connection.path} | {connection.client} |{self.ConnectedConnectionTime}| {connection.state}')
        connection = await self.dispatch(connection=connection)

        connection_time = perf_counter() - start_time  # Seconds
        logger.info(f'WS | {connection.path} | {connection.client} | {connection_time} | {connection.state}')
        return connection



from collections import defaultdict
from types import NoneType
from typing import Any

from pydantic import BaseModel

from panther.configs import config
from panther.response import IterableDataTypes


def _ref_name(ref: str) -> str:
    obj_name = ref.rsplit('/', maxsplit=1)[1]
    return f'${obj_name}'


def clean_model_schema(schema: dict) -> dict:
    """
    Example:
        {
        'title': 'Author',
        '$': {
            'Book': {
                'title': 'Book',
                'fields': {
                    'title': {'title': 'Title', 'type': ['string'], 'required': True},
                    'pages_count': {'title': 'Pages Count', 'type': ['integer'], 'required': True},
                    'readers': {'title': 'Readers', 'type': ['array', 'null'], 'items': '$Person', 'default': None, 'required': False},
                    'co_owner': {'type': ['$Person', 'null'], 'default': None, 'required': False}
                }
            },
            'Parent': {
                'title': 'Parent',
                'fields': {
                    'name': {'title': 'Name', 'type': ['string'], 'required': True},
                    'age': {'title': 'Age', 'type': ['string'], 'required': True},
                    'has_child': {'title': 'Has Child', 'type': ['boolean'], 'required': True}
                }
            },
            'Person': {
                'title': 'Person',
                'fields': {
                    'age': {'title': 'Age', 'type': ['integer'], 'required': True},
                    'real_name': {'title': 'Real Name', 'type': ['string'], 'required': True},
                    'parent': {'type': '$Parent', 'required': True},
                    'is_alive': {'title': 'Is Alive', 'type': ['boolean'], 'required': True},
                    'friends': {'title': 'Friends', 'type': ['array'], 'items': '$Person', 'required': True}
                }
            }
        },
        'fields': {
            '_id': {'title': ' Id', 'type': ['string', 'null'], 'default': None, 'required': False},
            'name': {'title': 'Name', 'type': ['string'], 'required': True},
            'person': {'type': ['$Person', 'null'], 'default': None, 'required': False},
            'books': {'title': 'Books', 'type': ['array'], 'items': '$Book', 'required': True},
            'is_male': {'title': 'Is Male', 'type': ['boolean', 'null'], 'required': True}
        }
    }

    """

    result = defaultdict(dict)
    result['title'] = schema['title']
    if '$defs' in schema:
        for sk, sv in schema['$defs'].items():
            result['$'][sk] = clean_model_schema(sv)

    for k, v in schema['properties'].items():
        result['fields'][k] = {}
        if 'title' in v:
            result['fields'][k]['title'] = v['title']

        if 'type' in v:
            result['fields'][k]['type'] = [v['type']]

        if 'anyOf' in v:
            result['fields'][k]['type'] = [i['type'] if 'type' in i else _ref_name(i['$ref']) for i in v['anyOf']]
            if 'array' in result['fields'][k]['type']:
                # One of them was array, so add the `items` field
                for t in v['anyOf']:
                    if 'items' in t:
                        result['fields'][k]['items'] = _ref_name(t['items']['$ref'])

        if 'default' in v:
            result['fields'][k]['default'] = v['default']

        if '$ref' in v:  # For obj
            result['fields'][k]['type'] = _ref_name(v['$ref'])

        if 'items' in v:  # For array
            result['fields'][k]['items'] = _ref_name(v['items']['$ref'])

        result['fields'][k]['required'] = k in schema.get('required', [])

    # Cast it to have a more clear stdout
    return dict(result)


def get_models():
    return [
        {
            'index': i,
            'name': model.__name__,
            'module': model.__module__,
        }
        for i, model in enumerate(config.MODELS)
    ]


def serialize_models(data: Any):
    if issubclass(type(data), BaseModel):
        return data.model_dump()

    elif isinstance(data, IterableDataTypes):
        return [serialize_models(d) for d in data]



from __future__ import annotations

from collections.abc import Iterable
from sys import version_info

from pantherdb import Cursor

from panther.db.connections import db
from panther.db.cursor import Cursor
from panther.db.queries.base_queries import BaseQuery
from panther.db.utils import prepare_id_for_query
from panther.exceptions import DatabaseError

if version_info >= (3, 11):
    from typing import Self
else:
    from typing import TypeVar

    Self = TypeVar('Self', bound='BasePantherDBQuery')


class BasePantherDBQuery(BaseQuery):
    @classmethod
    def _merge(cls, *args, is_mongo: bool = False) -> dict:
        return super()._merge(*args, is_mongo=is_mongo)

    # # # # # Find # # # # #
    @classmethod
    async def find_one(cls, _filter: dict | None = None, /, **kwargs) -> Self | None:
        if document := db.session.collection(cls.__name__).find_one(**cls._merge(_filter, kwargs)):
            return await cls._create_model_instance(document=document)
        return None

    @classmethod
    async def find(cls, _filter: dict | None = None, /, **kwargs) -> Cursor:
        cursor = db.session.collection(cls.__name__).find(**cls._merge(_filter, kwargs))
        cursor.response_type = cls._create_model_instance
        cursor.cls = cls
        return cursor

    @classmethod
    async def first(cls, _filter: dict | None = None, /, **kwargs) -> Self | None:
        if document := db.session.collection(cls.__name__).first(**cls._merge(_filter, kwargs)):
            return await cls._create_model_instance(document=document)
        return None

    @classmethod
    async def last(cls, _filter: dict | None = None, /, **kwargs) -> Self | None:
        if document := db.session.collection(cls.__name__).last(**cls._merge(_filter, kwargs)):
            return await cls._create_model_instance(document=document)
        return None

    @classmethod
    async def aggregate(cls, *args, **kwargs):
        msg = 'aggregate() does not supported in `PantherDB`.'
        raise DatabaseError(msg) from None

    # # # # # Count # # # # #
    @classmethod
    async def count(cls, _filter: dict | None = None, /, **kwargs) -> int:
        return db.session.collection(cls.__name__).count(**cls._merge(_filter, kwargs))

    # # # # # Insert # # # # #
    @classmethod
    async def insert_one(cls, document: dict) -> Self:
        insert_one_result = db.session.collection(cls.__name__).insert_one(**document)
        return insert_one_result['_id']

    @classmethod
    async def insert_many(cls, documents: Iterable[dict]) -> list[Self]:
        return [await cls.insert_one(document) for document in documents]

    # # # # # Delete # # # # #
    @classmethod
    async def delete_one(cls, _filter: dict | None = None, /, **kwargs) -> bool:
        return db.session.collection(cls.__name__).delete_one(**cls._merge(_filter, kwargs))

    @classmethod
    async def delete_many(cls, _filter: dict | None = None, /, **kwargs) -> int:
        return db.session.collection(cls.__name__).delete_many(**cls._merge(_filter, kwargs))

    # # # # # Update # # # # #
    @classmethod
    async def update_one(cls, _filter: dict, _update: dict | None = None, /, **kwargs) -> bool:
        prepare_id_for_query(_filter)

        # Step 1: Merge document parameters
        # Combine the _update dict with keyword arguments into a single document
        document = cls._merge(_update, kwargs)

        # Step 2: Process and validate document
        # - Validate data types and structure
        # - Convert Model instances to their IDs for database storage
        # - Handle File objects by saving to disk and storing file paths
        # - Process nested objects and relationships
        final_document = await cls._process_document(document)

        # Step 3: Create model instance (Validating)
        # - Retrieve Model instances from database using IDs
        # - Open File objects from their stored paths
        # - Build the complete model instance with all relationships
        await cls._create_model_instance(document=final_document, is_updating=True)

        return db.session.collection(cls.__name__).update_one(_filter, **final_document)

    @classmethod
    async def update_many(cls, _filter: dict, _update: dict | None = None, /, **kwargs) -> int:
        prepare_id_for_query(_filter)

        # Step 1: Merge document parameters
        # Combine the _update dict with keyword arguments into a single document
        document = cls._merge(_update, kwargs)

        # Step 2: Process and validate document
        # - Validate data types and structure
        # - Convert Model instances to their IDs for database storage
        # - Handle File objects by saving to disk and storing file paths
        # - Process nested objects and relationships
        final_document = await cls._process_document(document)

        # Step 3: Create model instance (Validating)
        # - Retrieve Model instances from database using IDs
        # - Open File objects from their stored paths
        # - Build the complete model instance with all relationships
        await cls._create_model_instance(document=final_document, is_updating=True)

        return db.session.collection(cls.__name__).update_many(_filter, **final_document)



from unittest import IsolatedAsyncioTestCase

from pydantic import BaseModel

from panther import Panther, status
from panther.app import API, GenericAPI
from panther.configs import config
from panther.db import Model
from panther.openapi.urls import url_routing
from panther.openapi.utils import EndpointParser
from panther.response import Response
from panther.test import APIClient


@API()
def my_api1():
    return {'detail': 'ok'}


@API()
def my_api2():
    data = {'detail': 'ok'}
    return data


@API()
def my_api3():
    my_data = {'detail': 'ok'}
    return Response(data=my_data)


@API()
async def my_api4():
    return Response(data={'detail': 'ok'})


@API()
async def my_api5():
    return Response(status_code=201)


@API()
async def my_api6():
    return Response(status_code=status.HTTP_207_MULTI_STATUS)


@API()
def my_api7():
    my_status = status.HTTP_207_MULTI_STATUS
    return Response(status_code=my_status)


@API()
def my_api8():
    my_status = 207
    return Response(status_code=my_status)


@API()
def my_api9():
    return Response({'detail': 'ok'})


@API()
def my_api10():
    my_data = {'detail': 'ok'}
    return Response(my_data)


@API()
def my_api11():
    my_data = {'detail': 'ok'}
    return Response(my_data, status_code=207)


@API()
def my_api12():
    my_data = {'detail': 'ok'}
    return Response(my_data, status_code=status.HTTP_207_MULTI_STATUS)


@API()
def my_api13():
    my_data = {'detail': 'ok'}
    my_status = 207
    return Response(my_data, status_code=my_status)


@API()
def my_api14():
    my_data = {'detail': 'ok'}
    my_status = status.HTTP_207_MULTI_STATUS
    return Response(my_data, status_code=my_status)


@API()
def my_api15():
    my_data = {'detail': 'ok'}
    my_status = status.HTTP_207_MULTI_STATUS
    return Response(my_data, my_status)


@API()
def my_api16():
    my_data = {'detail': 'ok'}
    return Response(my_data, status.HTTP_207_MULTI_STATUS)


@API()
def my_api17():
    return Response({}, 207)


class API18(GenericAPI):
    def get(self, *args, **kwargs):
        return {'detail': 'ok'}

    def post(self, *args, **kwargs):
        my_data = {'detail': 'ok'}
        return my_data

    async def put(self, *args, **kwargs):
        my_data = {'detail': 'ok'}
        return Response(data=my_data)

    def patch(self, *args, **kwargs):
        my_data = {'detail': 'ok'}
        return Response(data=my_data, status_code=201)

    def delete(self, *args, **kwargs):
        return Response({}, status.HTTP_204_NO_CONTENT)


@API()
def my_api19():
    my_response = ['1', '2']
    return my_response


@API()
def my_api20():
    my_response = 'Hello World!'
    return my_response


@API()
def my_api21():
    my_response = True
    return my_response


@API()
def my_api22():
    my_response = (6, 9)
    return my_response


@API()
def my_api23():
    my_response = Response(data={'detail': 'Hello'}, status_code=207)
    return my_response


class CustomBaseModel(BaseModel):
    name: str
    children: list[str] | None = None


@API()
def my_api24():
    my_response = CustomBaseModel(name='Ali', children=['A', 'B', 'C'])
    return my_response


class CustomModel(Model):
    title: str


@API()
def my_api25():
    my_response = CustomModel(title='Book')
    return my_response


@API()
def my_api26():
    return CustomBaseModel(name='Ali')


@API()
def my_api27():
    return CustomModel(title='Book')


class TestOpenAPI(IsolatedAsyncioTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        app = Panther(__name__, configs=__name__, urls={'docs': url_routing})
        cls.client = APIClient(app=app)

    @classmethod
    def tearDownClass(cls) -> None:
        config.refresh()

    # TODO: Uncomment and improve it at end of this pr

    async def test_swagger(self):
        response = await self.client.get('/docs/swagger/')
        expected_response = """<!doctype html>
<html>
  <head>
    <title>Swagger UI</title>
    <meta charset="utf-8" />
    <meta name="viewport" content="width=device-width, initial-scale=1" />
    <link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/swagger-ui-dist/swagger-ui.css" />
  </head>
  <body>
    <div id="swagger-ui"></div>
    <script src="https://cdn.jsdelivr.net/npm/swagger-ui-dist/swagger-ui-bundle.js"></script>
    <script>
      SwaggerUIBundle({ spec: {"components": {"schemas": {}, "securitySchemes": {"BearerAuth": {"bearerFormat": "JWT", "scheme": "bearer", "type": "http"}}}, "info": {"description": "Auto-generated OpenAPI documentation for Panther project.", "title": "Panther API", "version": "1.0.0"}, "openapi": "3.0.0", "paths": {"/docs/rapidoc/": {}, "/docs/redoc/": {}, "/docs/scalar/": {}, "/docs/spotlight/": {}, "/docs/swagger/": {}}, "security": [{"BearerAuth": []}]}, dom_id: '#swagger-ui' });
    </script>
  </body>
</html>"""
        assert expected_response == response.data

    async def test_scalar(self):
        response = await self.client.get('/docs/scalar/')
        expected_response = """<!doctype html>
<html>
  <head>
    <title>Scalar API Reference</title>
    <meta charset="utf-8" />
    <meta
      name="viewport"
      content="width=device-width, initial-scale=1" />
  </head>
  <body>
    <div id="app"></div>
    <script src="https://cdn.jsdelivr.net/npm/@scalar/api-reference"></script>
    <script>
      Scalar.createApiReference('#app', {
        content: {"components": {"schemas": {}, "securitySchemes": {"BearerAuth": {"bearerFormat": "JWT", "scheme": "bearer", "type": "http"}}}, "info": {"description": "Auto-generated OpenAPI documentation for Panther project.", "title": "Panther API", "version": "1.0.0"}, "openapi": "3.0.0", "paths": {"/docs/rapidoc/": {}, "/docs/redoc/": {}, "/docs/scalar/": {}, "/docs/spotlight/": {}, "/docs/swagger/": {}}, "security": [{"BearerAuth": []}]}
      })
    </script>
  </body>
</html>"""
        assert expected_response == response.data

    async def test_redoc(self):
        response = await self.client.get('/docs/redoc/')
        expected_response = """<!doctype html>
<html>
  <head>
    <title>ReDoc API Reference</title>
    <meta charset="utf-8" />
    <meta name="viewport" content="width=device-width, initial-scale=1" />
    <style>
      body {
        margin: 0;
        padding: 0;
      }
    </style>
  </head>
  <body>
    <div id="redoc-container"></div>
    <script src="https://cdn.jsdelivr.net/npm/redoc@next/bundles/redoc.standalone.js"></script>
    <script>
      Redoc.init({"components": {"schemas": {}, "securitySchemes": {"BearerAuth": {"bearerFormat": "JWT", "scheme": "bearer", "type": "http"}}}, "info": {"description": "Auto-generated OpenAPI documentation for Panther project.", "title": "Panther API", "version": "1.0.0"}, "openapi": "3.0.0", "paths": {"/docs/rapidoc/": {}, "/docs/redoc/": {}, "/docs/scalar/": {}, "/docs/spotlight/": {}, "/docs/swagger/": {}}, "security": [{"BearerAuth": []}]}, {}, document.getElementById("redoc-container"));
    </script>
  </body>
</html>"""
        assert expected_response == response.data

    async def test_rapidoc(self):
        response = await self.client.get('/docs/rapidoc/')
        expected_response = """<!doctype html>
<html>
  <head>
    <title>RapiDoc Inline JSON</title>
    <meta charset="utf-8">
    <meta name="viewport" content="width=device-width, minimum-scale=1, initial-scale=1, user-scalable=yes">
    <script type="module" src="https://unpkg.com/rapidoc/dist/rapidoc-min.js"></script>
  </head>
  <body>
    <rapi-doc id="thedoc"
      render-style = "read"
      allow-try="false"
      allow-authentication = "false"
    > </rapi-doc>
    <script>
      document.addEventListener('readystatechange', (event) => {
        let docEl = document.getElementById("thedoc");
        docEl.loadSpec({"components": {"schemas": {}, "securitySchemes": {"BearerAuth": {"bearerFormat": "JWT", "scheme": "bearer", "type": "http"}}}, "info": {"description": "Auto-generated OpenAPI documentation for Panther project.", "title": "Panther API", "version": "1.0.0"}, "openapi": "3.0.0", "paths": {"/docs/rapidoc/": {}, "/docs/redoc/": {}, "/docs/scalar/": {}, "/docs/spotlight/": {}, "/docs/swagger/": {}}, "security": [{"BearerAuth": []}]});
      })
    </script>
  </body>
</html>"""
        assert expected_response == response.data

    async def test_spotlight(self):
        response = await self.client.get('/docs/spotlight/')
        expected_response = """<!doctype html>
<html>
<head>
  <title>Stoplight Elements with CSS</title>
  <link rel="stylesheet" href="https://unpkg.com/@stoplight/elements/styles.min.css" />
  <script type="module" src="https://unpkg.com/@stoplight/elements/web-components.min.js"></script>
  <style>
    body, html {
      margin: 0; padding: 0; height: 100%;
    }
    elements-api {
      display: block;
      height: 100vh;
    }
  </style>
</head>
<body>
  <elements-api id="api-viewer" router="hash"></elements-api>
  <script>
    customElements.whenDefined("elements-api").then(() => {
      document.getElementById("api-viewer").apiDescriptionDocument = {"components": {"schemas": {}, "securitySchemes": {"BearerAuth": {"bearerFormat": "JWT", "scheme": "bearer", "type": "http"}}}, "info": {"description": "Auto-generated OpenAPI documentation for Panther project.", "title": "Panther API", "version": "1.0.0"}, "openapi": "3.0.0", "paths": {"/docs/rapidoc/": {}, "/docs/redoc/": {}, "/docs/scalar/": {}, "/docs/spotlight/": {}, "/docs/swagger/": {}}, "security": [{"BearerAuth": []}]};
    });
  </script>
</body>
</html>"""
        assert expected_response == response.data

    async def test_my_api1(self):
        parsed = EndpointParser(my_api1, 'get')
        assert parsed.status_code == 200
        assert parsed.response_data == {'detail': 'ok'}

    async def test_my_api2(self):
        parsed = EndpointParser(my_api2, 'get')
        assert parsed.status_code == 200
        assert parsed.response_data == {'detail': 'ok'}

    async def test_my_api3(self):
        parsed = EndpointParser(my_api3, 'get')
        assert parsed.status_code == 200
        assert parsed.response_data == {'detail': 'ok'}

    async def test_my_api4(self):
        parsed = EndpointParser(my_api4, 'get')
        assert parsed.status_code == 200
        assert parsed.response_data == {'detail': 'ok'}

    async def test_my_api5(self):
        parsed = EndpointParser(my_api5, 'get')
        assert parsed.status_code == 201
        assert parsed.response_data == {}

    async def test_my_api6(self):
        parsed = EndpointParser(my_api6, 'get')
        assert parsed.status_code == 207
        assert parsed.response_data == {}

    async def test_my_api7(self):
        parsed = EndpointParser(my_api7, 'get')
        assert parsed.status_code == 207
        assert parsed.response_data == {}

    async def test_my_api8(self):
        parsed = EndpointParser(my_api8, 'get')
        assert parsed.status_code == 207
        assert parsed.response_data == {}

    async def test_my_api9(self):
        parsed = EndpointParser(my_api9, 'get')
        assert parsed.status_code == 200
        assert parsed.response_data == {'detail': 'ok'}

    async def test_my_api10(self):
        parsed = EndpointParser(my_api10, 'get')
        assert parsed.status_code == 200
        assert parsed.response_data == {'detail': 'ok'}

    async def test_my_api11(self):
        parsed = EndpointParser(my_api11, 'get')
        assert parsed.status_code == 207
        assert parsed.response_data == {'detail': 'ok'}

    async def test_my_api12(self):
        parsed = EndpointParser(my_api12, 'get')
        assert parsed.status_code == 207
        assert parsed.response_data == {'detail': 'ok'}

    async def test_my_api13(self):
        parsed = EndpointParser(my_api13, 'get')
        assert parsed.status_code == 207
        assert parsed.response_data == {'detail': 'ok'}

    async def test_my_api14(self):
        parsed = EndpointParser(my_api14, 'get')
        assert parsed.status_code == 207
        assert parsed.response_data == {'detail': 'ok'}

    async def test_my_api15(self):
        parsed = EndpointParser(my_api15, 'get')
        assert parsed.status_code == 207
        assert parsed.response_data == {'detail': 'ok'}

    async def test_my_api16(self):
        parsed = EndpointParser(my_api16, 'get')
        assert parsed.status_code == 207
        assert parsed.response_data == {'detail': 'ok'}

    async def test_my_api17(self):
        parsed = EndpointParser(my_api17, 'get')
        assert parsed.status_code == 207
        assert parsed.response_data == {}

    async def test_my_api18_get(self):
        parsed = EndpointParser(API18, 'get')
        assert parsed.status_code == 200
        assert parsed.response_data == {'detail': 'ok'}

    async def test_my_api18_post(self):
        parsed = EndpointParser(API18, 'post')
        assert parsed.status_code == 200
        assert parsed.response_data == {'detail': 'ok'}

    async def test_my_api18_put(self):
        parsed = EndpointParser(API18, 'put')
        assert parsed.status_code == 200
        assert parsed.response_data == {'detail': 'ok'}

    async def test_my_api18_patch(self):
        parsed = EndpointParser(API18, 'patch')
        assert parsed.status_code == 201
        assert parsed.response_data == {'detail': 'ok'}

    async def test_my_api18_delete(self):
        parsed = EndpointParser(API18, 'delete')
        assert parsed.status_code == 204
        assert parsed.response_data == {}

    async def test_my_api19(self):
        parsed = EndpointParser(my_api19, 'get')
        assert parsed.status_code == 200
        assert parsed.response_data == ['1', '2']

    async def test_my_api20(self):
        parsed = EndpointParser(my_api20, 'get')
        assert parsed.status_code == 200
        assert parsed.response_data == 'Hello World!'

    async def test_my_api21(self):
        parsed = EndpointParser(my_api21, 'get')
        assert parsed.status_code == 200
        assert parsed.response_data == True

    async def test_my_api22(self):
        parsed = EndpointParser(my_api22, 'get')
        assert parsed.status_code == 200
        assert parsed.response_data == [6, 9]

    async def test_my_api23(self):
        parsed = EndpointParser(my_api23, 'get')
        assert parsed.status_code == 207
        assert parsed.response_data == {'detail': 'Hello'}

    async def test_my_api24(self):
        parsed = EndpointParser(my_api24, 'get')
        assert parsed.status_code == 200
        assert parsed.response_data == {'name': 'Ali', 'children': ['A', 'B', 'C']}

    async def test_my_api25(self):
        parsed = EndpointParser(my_api25, 'get')
        assert parsed.status_code == 200
        assert parsed.response_data == {'title': 'Book'}

    async def test_my_api26(self):
        parsed = EndpointParser(my_api26, 'get')
        assert parsed.status_code == 200
        assert parsed.response_data == {'name': 'Ali'}

    async def test_my_api27(self):
        parsed = EndpointParser(my_api27, 'get')
        assert parsed.status_code == 200
        assert parsed.response_data == {'title': 'Book'}



import logging
import sys
from collections.abc import Callable
from logging.config import dictConfig
from pathlib import Path

import panther.logging
from panther import status
from panther._load_configs import *
from panther._utils import ENDPOINT_CLASS_BASED_API, ENDPOINT_FUNCTION_BASED_API, reformat_code, traceback_message
from panther.base_websocket import Websocket
from panther.cli.utils import print_info
from panther.configs import config
from panther.events import Event
from panther.exceptions import APIError, BaseError, NotFoundAPIError, PantherError, UpgradeRequiredError
from panther.request import Request
from panther.response import Response
from panther.routings import find_endpoint
from panther.websocket import GenericWebsocket

dictConfig(panther.logging.LOGGING)
logger = logging.getLogger('panther')


class Panther:
    def __init__(self, name: str, configs: str | None = None, urls: dict | None = None):
        """
        Initialize a Panther application instance.

        Args:
            name: Typically set to `__name__`; used to determine the current directory of the application.
            configs: The name of the module containing your configuration.
                If the configuration is defined in the current file, you can also set this to `__name__`.
            urls: A dictionary containing your URL routing.
                If not provided, Panther will attempt to load `URLs` from the configs module.

        """
        self._configs_module_name = configs
        self._urls = urls

        config.BASE_DIR = Path(name).resolve().parent

        try:
            self.load_configs()
            if config.AUTO_REFORMAT:
                reformat_code(base_dir=config.BASE_DIR)
        except Exception as e:
            logger.error(e.args[0] if isinstance(e, PantherError) else traceback_message(exception=e))
            sys.exit()

        # Print Info
        print_info(config)

    def load_configs(self) -> None:
        # Check & Read The Configs File
        self._configs_module = load_configs_module(self._configs_module_name)

        load_redis(self._configs_module)
        load_timezone(self._configs_module)
        load_database(self._configs_module)
        load_secret_key(self._configs_module)
        load_throttling(self._configs_module)
        load_user_model(self._configs_module)
        load_log_queries(self._configs_module)
        load_templates_dir(self._configs_module)
        load_middlewares(self._configs_module)
        load_auto_reformat(self._configs_module)
        load_background_tasks(self._configs_module)
        load_other_configs(self._configs_module)
        load_urls(self._configs_module, urls=self._urls)
        load_authentication_class(self._configs_module)
        load_websocket_connections()

        check_endpoints_inheritance()

    async def __call__(self, scope: dict, receive: Callable, send: Callable) -> None:
        if scope['type'] == 'http':
            await self.handle_http(scope=scope, receive=receive, send=send)
        elif scope['type'] == 'websocket':
            await self.handle_ws(scope=scope, receive=receive, send=send)
        elif scope['type'] == 'lifespan':
            message = await receive()
            if message['type'] == 'lifespan.startup':
                if config.HAS_WS:
                    await config.WEBSOCKET_CONNECTIONS.start()
                await Event.run_startups()
            elif message['type'] == 'lifespan.shutdown':
                pass
            return

    @staticmethod
    async def handle_ws_endpoint(connection: Websocket):
        # Find Endpoint
        endpoint, found_path = find_endpoint(path=connection.path)
        if endpoint is None:
            logger.debug(f'Path `{connection.path}` not found')
            await connection.close()
            return connection

        # Check Endpoint Type
        if not issubclass(endpoint, GenericWebsocket):
            logger.warning(f'{endpoint.__name__}() class is not a Websocket class.')
            await connection.close()
            return connection

        # Create The Connection
        final_connection = endpoint(parent=connection)
        del connection

        # Collect Path Variables
        final_connection.collect_path_variables(found_path=found_path)

        return await config.WEBSOCKET_CONNECTIONS.listen(connection=final_connection)

    @classmethod
    async def handle_ws(cls, scope: dict, receive: Callable, send: Callable) -> None:
        # Create Temp Connection
        connection = Websocket(scope=scope, receive=receive, send=send)

        # Create Middlewares chain
        chained_func = cls.handle_ws_endpoint
        for middleware in reversed(config.WS_MIDDLEWARES):
            chained_func = middleware(dispatch=chained_func)

        # Call Middlewares & Endpoint
        try:
            connection = await chained_func(connection=connection)
        except BaseError as e:
            connection.log(e.detail)
            await connection.close()
        except Exception as e:
            logger.error(traceback_message(exception=e))
            await connection.close()

    @staticmethod
    async def handle_http_endpoint(request: Request) -> Response:
        # Find Endpoint
        endpoint, found_path = find_endpoint(path=request.path)
        if endpoint is None:
            raise NotFoundAPIError

        # Collect Path Variables
        request.collect_path_variables(found_path=found_path)

        if endpoint._endpoint_type is ENDPOINT_FUNCTION_BASED_API:
            return await endpoint(request=request)
        if endpoint._endpoint_type is ENDPOINT_CLASS_BASED_API:
            return await endpoint().call_method(request=request)

        # ENDPOINT_WEBSOCKET
        raise UpgradeRequiredError

    @classmethod
    async def handle_http(cls, scope: dict, receive: Callable, send: Callable) -> None:
        # Create `Request` and its body
        request = Request(scope=scope, receive=receive, send=send)
        await request.read_body()

        # Create Middlewares chain
        chained_func = cls.handle_http_endpoint
        for middleware in reversed(config.HTTP_MIDDLEWARES):
            chained_func = middleware(dispatch=chained_func)

        # Call Middlewares & Endpoint
        try:
            response = await chained_func(request=request)
            if response is None:
                logger.error('You forgot to return `response` on the `Middlewares.__call__()`')
                response = Response(
                    data={'detail': 'Internal Server Error'},
                    status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                )
        # Handle `APIError` Exceptions
        except APIError as e:
            response = Response(
                data=e.detail if isinstance(e.detail, dict) else {'detail': e.detail},
                headers=e.headers,
                status_code=e.status_code,
            )
        # Handle Unknown Exceptions
        except Exception as e:
            logger.error(traceback_message(exception=e))
            response = Response(
                data={'detail': 'Internal Server Error'},
                status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            )

        # Return Response
        await response.send(send=send, receive=receive)

    def __del__(self):
        Event.run_shutdowns()



import logging
from collections import namedtuple
from datetime import datetime, timedelta

import orjson as json

from panther.db.connections import redis
from panther.request import Request
from panther.response import Response
from panther.utils import generate_hash_value_from_string, round_datetime

logger = logging.getLogger('panther')

caches: dict[str, tuple[bytes, dict, int]] = {}
CachedResponse = namedtuple('CachedResponse', ['data', 'headers', 'status_code'])


def api_cache_key(request: Request, duration: timedelta | None = None) -> str:
    client = (request.user and request.user.id) or request.client.ip
    query_params_hash = generate_hash_value_from_string(request.scope['query_string'].decode('utf-8'))
    key = f'{client}-{request.path}-{query_params_hash}-{request.validated_data}'

    if duration:
        time = round_datetime(datetime.now(), duration)
        return f'{time}-{key}'

    return key


async def get_response_from_cache(*, request: Request, duration: timedelta) -> CachedResponse | None:
    """
    If redis.is_connected:
        Get Cached Data From Redis
    else:
        Get Cached Data From Memory
    """
    if redis.is_connected:
        key = api_cache_key(request=request)
        data = (await redis.get(key) or b'{}').decode()
        if value := json.loads(data):
            return CachedResponse(data=value[0].encode(), headers=value[1], status_code=value[2])
    else:
        key = api_cache_key(request=request, duration=duration)
        if value := caches.get(key):
            return CachedResponse(*value)


async def set_response_in_cache(*, request: Request, response: Response, duration: timedelta | int) -> None:
    """
    If redis.is_connected:
        Cache The Data In Redis
    else:
        Cache The Data In Memory
    """

    if redis.is_connected:
        key = api_cache_key(request=request)
        cache_data: tuple[str, dict, int] = (response.body.decode(), response.headers, response.status_code)
        await redis.set(key, json.dumps(cache_data), ex=duration)

    else:
        key = api_cache_key(request=request, duration=duration)
        caches[key] = (response.body, response.headers, response.status_code)
        logger.info('`cache` is not very accurate when `redis` is not connected.')



import datetime
import typing
from typing import Any

from pydantic import BaseModel, BeforeValidator, create_model
from pydantic.fields import FieldInfo
from pydantic_core._pydantic_core import PydanticUndefined

from panther.db import Model
from panther.utils import run_coroutine, timezone_now


class MetaModelSerializer:
    KNOWN_CONFIGS = ['model', 'fields', 'exclude', 'required_fields', 'optional_fields']

    def __new__(cls, cls_name: str, bases: tuple[type[typing.Any], ...], namespace: dict[str, typing.Any], **kwargs):
        if cls_name == 'ModelSerializer':
            # Put `model` to the main class with `create_model()`
            namespace['__annotations__'].pop('model')
            cls.model_serializer = type(cls_name, (), namespace)
            return super().__new__(cls)

        # 1. Initial Check
        cls.check_config(cls_name=cls_name, namespace=namespace)
        config = namespace.pop('Config')

        # 2. Collect `Fields`
        field_definitions = cls.collect_fields(config=config, namespace=namespace)

        # 3. Collect `pydantic.model_config`
        model_config = cls.collect_model_config(config=config, namespace=namespace)
        namespace |= {'model_config': model_config}
        # 4. Create a serializer
        return create_model(
            cls_name,
            __module__=namespace['__module__'],
            __validators__=namespace,
            __base__=(cls.model_serializer, BaseModel),
            model=(typing.ClassVar[type[BaseModel]], config.model),
            **field_definitions,
        )

    @classmethod
    def check_config(cls, cls_name: str, namespace: dict) -> None:
        module = namespace['__module__']
        address = f'{module}.{cls_name}'

        # Check `Config`
        if (config := namespace.get('Config')) is None:
            msg = f'`class Config` is required in {address}.'
            raise AttributeError(msg) from None

        # Check `model`
        if (model := getattr(config, 'model', None)) is None:
            msg = f'`{cls_name}.Config.model` is required.'
            raise AttributeError(msg) from None

        # Check `model` type
        try:
            if not issubclass(model, (Model, BaseModel)):
                msg = f'`{cls_name}.Config.model` is not subclass of `panther.db.Model` or `pydantic.BaseModel`.'
                raise AttributeError(msg) from None
        except TypeError:
            msg = f'`{cls_name}.Config.model` is not subclass of `panther.db.Model`.'
            raise AttributeError(msg) from None

        # Check `fields`
        if not hasattr(config, 'fields'):
            msg = f'`{cls_name}.Config.fields` is required.'
            raise AttributeError(msg) from None

        if config.fields != '*':
            for field_name in config.fields:
                if field_name == '*':
                    msg = f"`{cls_name}.Config.fields.{field_name}` is not valid. Did you mean `fields = '*'`"
                    raise AttributeError(msg) from None

                if field_name not in model.model_fields:
                    msg = f'`{cls_name}.Config.fields.{field_name}` is not in `{model.__name__}.model_fields`'
                    raise AttributeError(msg) from None

        # Check `required_fields`
        if not hasattr(config, 'required_fields'):
            config.required_fields = []

        for required in config.required_fields:
            if required not in model.model_fields:
                msg = f'`{cls_name}.Config.required_fields.{required}` is not valid.'
                raise AttributeError(msg) from None

            if config.fields != '*' and required not in config.fields:
                msg = f'`{cls_name}.Config.required_fields.{required}` is not defined in `Config.fields`.'
                raise AttributeError(msg) from None

        # Check `optional_fields`
        if not hasattr(config, 'optional_fields'):
            config.optional_fields = []

        for optional in config.optional_fields:
            if optional not in model.model_fields:
                msg = f'`{cls_name}.Config.optional_fields.{optional}` is not valid.'
                raise AttributeError(msg) from None

            if config.fields != '*' and optional not in config.fields:
                msg = f'`{cls_name}.Config.optional_fields.{optional}` is not defined in `Config.fields`.'
                raise AttributeError(msg) from None

        # Check `required_fields` and `optional_fields` together
        if (config.optional_fields == '*' and config.required_fields != []) or (
            config.required_fields == '*' and config.optional_fields != []
        ):
            msg = (
                f'`{cls_name}.Config.optional_fields` and '
                f"`{cls_name}.Config.required_fields` can't include same fields at the same time"
            )
            raise AttributeError(msg) from None
        for optional in config.optional_fields:
            for required in config.required_fields:
                if optional == required:
                    msg = (
                        f"`{optional}` can't be in `{cls_name}.Config.optional_fields` and "
                        f'`{cls_name}.Config.required_fields` at the same time'
                    )
                    raise AttributeError(msg) from None

        # Check `exclude`
        if not hasattr(config, 'exclude'):
            config.exclude = []

        for field_name in config.exclude:
            if field_name not in model.model_fields:
                msg = f'`{cls_name}.Config.exclude.{field_name}` is not valid.'
                raise AttributeError(msg) from None

            if config.fields != '*' and field_name not in config.fields:
                msg = f'`{cls_name}.Config.exclude.{field_name}` is not defined in `Config.fields`.'
                raise AttributeError(msg) from None

    @classmethod
    def collect_fields(cls, config: typing.Callable, namespace: dict) -> dict:
        field_definitions = {}

        # Define `fields`
        if config.fields == '*':
            for field_name, field in config.model.model_fields.items():
                field_definitions[field_name] = (field.annotation, field)
        else:
            for field_name in config.fields:
                field_definitions[field_name] = (
                    config.model.model_fields[field_name].annotation,
                    config.model.model_fields[field_name],
                )

        # Apply `exclude`
        for field_name in config.exclude:
            del field_definitions[field_name]

        # Apply `required_fields`
        if config.required_fields == '*':
            for value in field_definitions.values():
                value[1].default = PydanticUndefined
        else:
            for field_name in config.required_fields:
                field_definitions[field_name][1].default = PydanticUndefined

        # Apply `optional_fields`
        if config.optional_fields == '*':
            for value in field_definitions.values():
                value[1].default = value[0]()
        else:
            for field_name in config.optional_fields:
                if issubclass(field_definitions[field_name][0], datetime.datetime):
                    field_definitions[field_name][1].default = timezone_now()
                else:
                    field_definitions[field_name][1].default = field_definitions[field_name][0]()
        # Collect and Override `Class Fields`
        for key, value in namespace.pop('__annotations__', {}).items():
            field_info = namespace.pop(key, FieldInfo(annotation=value))
            field_definitions[key] = (value, field_info)

        # Check Foreign Keys
        for field_name, field_config in field_definitions.items():
            try:
                if issubclass(field_config[0], Model):
                    validator = BeforeValidator(cls.convert_str_to_model(field_config[0]))
                    ann = typing.Annotated[field_config[0], validator]
                    field_definitions[field_name] = (ann, field_config[1])
            except TypeError:
                pass
        return field_definitions

    @classmethod
    def collect_model_config(cls, config: typing.Callable, namespace: dict) -> dict:
        return (
            {
                attr: getattr(config, attr)
                for attr in dir(config)
                if not attr.startswith('__') and attr not in cls.KNOWN_CONFIGS
            }
            | namespace.pop('model_config', {})
            | {'arbitrary_types_allowed': True}
        )

    @classmethod
    def convert_str_to_model(cls, model_cls):
        def _convert(v: Any) -> Any:
            if isinstance(v, str):
                return run_coroutine(model_cls.find_one(id=v))
            return v

        return _convert


class ModelSerializer(metaclass=MetaModelSerializer):
    """
    A serializer class that automatically generates Pydantic models from database models.

    Documentation:
        https://pantherpy.github.io/serializer/#style-2-modelserializer

    Example:
        class PersonSerializer(ModelSerializer):
            class Config:
                # Required: The model class to serialize
                model = Person

                # Required: Fields to include in serialization
                # Use '*' for all fields, or specify a list of field names
                fields = '*'  # or ['id', 'name', 'email']

                # Optional: Fields to exclude from serialization
                # Must be a list of field names (cannot be '*')
                exclude = ['created_date', 'updated_date']

                # Optional: Fields that are required (no default value)
                # Can be a list of field names or '*' for all fields
                required_fields = ['first_name', 'last_name']

                # Optional: Fields that are optional (with default values)
                # Can be a list of field names or '*' for all fields
                optional_fields = ['age', 'bio']
    """

    model: type[BaseModel]

    async def to_response(self, instance: Any, data: dict) -> dict:
        return data



import re
import types
from collections import Counter
from collections.abc import Callable, Mapping, MutableMapping
from copy import deepcopy
from functools import partial, reduce

from panther.configs import config
from panther.exceptions import PantherError


def flatten_urls(urls: dict) -> dict:
    return dict(_flattening_urls(urls))


def _flattening_urls(data: dict | Callable, url: str = ''):
    # Add `/` add the end of url
    if not url.endswith('/'):
        url = f'{url}/'

    if isinstance(data, dict):
        if data == {}:
            # User didn't define any endpoint,
            #   So we just reserve this path so won't be used in path variables.
            yield url.removeprefix('/'), {}

        for k, v in data.items():
            yield from _flattening_urls(v, f'{url}{k}')
    else:
        # Remove `/` prefix of url
        url = url.removeprefix('/')

        # Collect it, if it doesn't have problem
        _is_url_endpoint_valid(url=url, endpoint=data)
        yield url, data


def _is_url_endpoint_valid(url: str, endpoint: Callable):
    if endpoint is ...:
        raise PantherError(f"URL Can't Point To Ellipsis. ('{url}' -> ...)")
    if endpoint is None:
        raise PantherError(f"URL Can't Point To None. ('{url}' -> None)")
    if url and not re.match(r'^[a-zA-Z<>0-9_/-]+$', url):
        raise PantherError(f"URL Is Not Valid. --> '{url}'")
    elif isinstance(endpoint, types.ModuleType):
        raise PantherError(f"URL Can't Point To Module. --> '{url}'")


def finalize_urls(urls: dict) -> dict:
    """Convert flat dict to nested"""
    urls_list = []
    for url, endpoint in urls.items():
        path = {}
        if url == '':
            # This condition only happen when
            #   user defines the root url == '' instead of '/'
            url = '/'

        for single_path in list(filter(lambda x: x != '', url.split('/')[:-1][::-1])) or ['']:
            if single_path != '' and not path:
                path = {single_path: {'': endpoint}}
            else:
                path = {single_path: path or endpoint}
        urls_list.append(path)
    final_urls = _merge(*urls_list) if urls_list else {}
    check_urls_path_variables(final_urls)
    return final_urls


def check_urls_path_variables(urls: dict, path: str = '') -> None:
    middle_route_error = []
    last_route_error = []
    for key, value in urls.items():
        new_path = f'{path}/{key}'

        if isinstance(value, dict):
            if key.startswith('<'):
                middle_route_error.append(new_path)
            check_urls_path_variables(value, path=new_path)
        elif key.startswith('<'):
            last_route_error.append(new_path)

    if len(middle_route_error) > 1:
        msg = '\n\t- ' + '\n\t- '.join(e for e in middle_route_error)
        raise PantherError(f"URLs can't have same-level path variables that point to a dict: {msg}")

    if len(last_route_error) > 1:
        msg = '\n\t- ' + '\n\t- '.join(e for e in last_route_error)
        raise PantherError(f"URLs can't have same-level path variables that point to an endpoint: {msg}")


def _merge(destination: MutableMapping, *sources) -> MutableMapping:
    return _simplify_urls(reduce(partial(_deepmerge), sources, destination))


def _simplify_urls(urls):
    simplified_urls = {}

    for key, value in urls.items():
        if isinstance(value, dict):
            simplified_value = _simplify_urls(value)
            if isinstance(simplified_value, dict) and len(simplified_value) == 1 and '' in simplified_value:
                simplified_urls[key] = simplified_value['']
            else:
                simplified_urls[key] = simplified_value
        else:
            simplified_urls[key] = value

    return simplified_urls


def _deepmerge(dst, src):
    """Credit to Travis Clarke --> https://github.com/clarketm/mergedeep"""
    for key in src:
        if key in dst:
            if _is_recursive_merge(dst[key], src[key]):
                _deepmerge(dst[key], src[key])
            else:
                dst[key] = deepcopy(src[key])
        else:
            dst[key] = deepcopy(src[key])
    return dst


def _is_recursive_merge(a, b):
    both_mapping = isinstance(a, Mapping) and isinstance(b, Mapping)
    both_counter = isinstance(a, Counter) and isinstance(b, Counter)
    return both_mapping and not both_counter


ENDPOINT_NOT_FOUND = (None, '')


def find_endpoint(path: str) -> tuple[Callable | None, str]:
    urls = config.URLS

    # 'user/list/?name=ali' --> 'user/list/' --> 'user/list' --> ['user', 'list']
    parts = path.split('?')[0].strip('/').split('/')
    paths_len = len(parts)

    found_path = []
    for i, part in enumerate(parts):
        last_path = bool((i + 1) == paths_len)
        found = urls.get(part)

        if last_path:
            # `found` is callable
            if callable(found):
                found_path.append(part)
                return found, '/'.join(found_path)

            # `found` is dict
            if isinstance(found, dict):
                if (endpoint := found.get('')) and callable(endpoint):
                    found_path.append(part)
                    return endpoint, '/'.join(found_path)
                else:
                    return ENDPOINT_NOT_FOUND

            # `found` is None
            for key, value in urls.items():
                if key.startswith('<'):
                    if callable(value):
                        found_path.append(key)
                        return value, '/'.join(found_path)

                    elif isinstance(value, dict) and (endpoint := value.get('')):
                        if callable(endpoint):
                            found_path.append(key)
                            return endpoint, '/'.join(found_path)
                        else:
                            return ENDPOINT_NOT_FOUND

            return ENDPOINT_NOT_FOUND

        # `found` is dict
        elif isinstance(found, dict):
            found_path.append(part)
            urls = found
            continue

        # `found` is callable
        elif callable(found):
            return ENDPOINT_NOT_FOUND

        else:
            # `found` is None
            for key, value in urls.items():
                if key.startswith('<') and isinstance(value, dict):
                    found_path.append(key)
                    urls = value
                    break
            else:
                return ENDPOINT_NOT_FOUND



import random
from unittest import TestCase

from panther.base_request import BaseRequest
from panther.configs import config
from panther.exceptions import PantherError
from panther.routings import (
    ENDPOINT_NOT_FOUND,
    finalize_urls,
    find_endpoint,
    flatten_urls,
)


class TestRoutingFunctions(TestCase):
    def tearDown(self) -> None:
        config.refresh()

    @classmethod
    def tearDownClass(cls):
        config.refresh()

    # Collecting
    def test_flatten_urls_ellipsis_endpoints(self):
        urls1 = {
            'user/': {
                '<user_id>/': ...,
            },
        }
        urls2 = {
            'user/': {
                'list/': ...,
            },
        }

        try:
            flatten_urls(urls1)
        except PantherError as exc:
            assert exc.args[0] == "URL Can't Point To Ellipsis. ('user/<user_id>/' -> ...)"
        else:
            assert False

        try:
            flatten_urls(urls2)
        except PantherError as exc:
            assert exc.args[0] == "URL Can't Point To Ellipsis. ('user/list/' -> ...)"
        else:
            assert False

    def test_flatten_urls_none_endpoints(self):
        urls1 = {
            'user/': {
                'list/': None,
            },
        }
        urls2 = {
            'user/': {
                '<user_id>/': None,
            },
        }

        try:
            flatten_urls(urls1)
        except PantherError as exc:
            assert exc.args[0] == "URL Can't Point To None. ('user/list/' -> None)"
        else:
            assert False

        try:
            flatten_urls(urls2)
        except PantherError as exc:
            assert exc.args[0] == "URL Can't Point To None. ('user/<user_id>/' -> None)"
        else:
            assert False

    def test_flatten_urls_invalid_url(self):
        def temp_func():
            pass

        urls1 = {
            'user/': {
                '?': temp_func,
            },
        }
        urls2 = {
            'user/': {
                'لیست': temp_func,
            },
        }

        try:
            flatten_urls(urls1)
        except PantherError as exc:
            assert exc.args[0] == "URL Is Not Valid. --> 'user/?/'"
        else:
            assert False

        try:
            flatten_urls(urls2)
        except PantherError as exc:
            assert exc.args[0] == "URL Is Not Valid. --> 'user/لیست/'"
        else:
            assert False

    def test_flatten_urls_empty_url(self):
        def temp_func():
            pass

        urls = {
            '': temp_func,
        }

        collected_urls = flatten_urls(urls)

        expected_result = {
            '': temp_func,
        }
        assert collected_urls == expected_result

    def test_flatten_urls_nested_empty_urls(self):
        def temp_func():
            pass

        urls = {
            '': {
                '': {
                    'nested/<user_id>/': temp_func,
                    'nested/profile/': temp_func,
                    'nested/list/': temp_func,
                },
                '<user_id>/': temp_func,
                'profile/': temp_func,
                'list/': temp_func,
            },
        }

        collected_urls = flatten_urls(urls)

        expected_result = {
            '<user_id>/': temp_func,
            'profile/': temp_func,
            'list/': temp_func,
            'nested/<user_id>/': temp_func,
            'nested/profile/': temp_func,
            'nested/list/': temp_func,
        }
        assert collected_urls == expected_result

    def test_flatten_urls_slash_url(self):
        def temp_func():
            pass

        urls = {
            '/': temp_func,
        }

        collected_urls = flatten_urls(urls)

        expected_result = {
            '/': temp_func,
        }
        assert collected_urls == expected_result

    def test_flatten_urls_nested_slash_urls(self):
        def temp_func():
            pass

        urls = {
            '/': {
                '/': {
                    'nested/<user_id>/': temp_func,
                    'nested/profile/': temp_func,
                    'nested/list/': temp_func,
                },
                '<user_id>/': temp_func,
                'profile/': temp_func,
                'list/': temp_func,
            },
        }

        collected_urls = flatten_urls(urls)

        expected_result = {
            '/<user_id>/': temp_func,
            '/profile/': temp_func,
            '/list/': temp_func,
            '//nested/<user_id>/': temp_func,
            '//nested/profile/': temp_func,
            '//nested/list/': temp_func,
        }
        assert collected_urls == expected_result

    def test_flatten_urls_simple_urls(self):
        def temp_func():
            pass

        urls = {
            '<user_id>/': temp_func,
            'profile/': temp_func,
            'list/': temp_func,
        }

        collected_urls = flatten_urls(urls)

        expected_result = {
            '<user_id>/': temp_func,
            'profile/': temp_func,
            'list/': temp_func,
        }
        assert collected_urls == expected_result

    def test_flatten_urls_simple_nested_urls(self):
        def temp_func():
            pass

        urls = {
            'user/': {
                '<user_id>/': temp_func,
                'profile/': temp_func,
                'list/': temp_func,
            },
        }

        collected_urls = flatten_urls(urls)

        expected_result = {
            'user/<user_id>/': temp_func,
            'user/profile/': temp_func,
            'user/list/': temp_func,
        }
        assert collected_urls == expected_result

    def test_flatten_urls_simple_nested_urls_without_slash_at_end(self):
        def temp_func():
            pass

        urls = {
            'user': {
                '<user_id>': temp_func,
                'profile': temp_func,
                'list': temp_func,
            },
        }

        collected_urls = flatten_urls(urls)

        expected_result = {
            'user/<user_id>/': temp_func,
            'user/profile/': temp_func,
            'user/list/': temp_func,
        }
        assert collected_urls == expected_result

    def test_flatten_urls_complex_nested_urls(self):
        def temp_func():
            pass

        urls = {
            'user/': {
                '<user_id>/profile/<id>': temp_func,
                'profile/': temp_func,
                'list/': temp_func,
            },
            '': {
                'payments': temp_func,
                'notifications': temp_func,
            },
            'admin/v1': {
                'profile/avatar': temp_func,
                '<user_id>': temp_func,
                'users/': {
                    'list': {
                        'registered': temp_func,
                        'not-registered': temp_func,
                    },
                    'detail': {
                        'registered': temp_func,
                        'not-registered': temp_func,
                    },
                },
            },
            'admin/v2': {},
        }

        collected_urls = flatten_urls(urls)
        expected_result = {
            'user/<user_id>/profile/<id>/': temp_func,
            'user/profile/': temp_func,
            'user/list/': temp_func,
            'payments/': temp_func,
            'notifications/': temp_func,
            'admin/v1/profile/avatar/': temp_func,
            'admin/v1/<user_id>/': temp_func,
            'admin/v1/users/list/registered/': temp_func,
            'admin/v1/users/list/not-registered/': temp_func,
            'admin/v1/users/detail/registered/': temp_func,
            'admin/v1/users/detail/not-registered/': temp_func,
            'admin/v2/': {},
        }
        assert collected_urls == expected_result

    # Finalize
    def test_finalize_empty_url(self):
        def temp_func():
            pass

        urls = {
            '': temp_func,
        }

        collected_urls = flatten_urls(urls)
        finalized_urls = finalize_urls(collected_urls)

        expected_result = {
            '': temp_func,
        }
        assert finalized_urls == expected_result

    def test_finalize_empty_url_nested(self):
        def temp_func():
            pass

        urls = {
            '': {
                '': {
                    'nested/<user_id>/': temp_func,
                    'nested/profile/': temp_func,
                    'nested/list/': temp_func,
                },
                '<user_id>/': temp_func,
                'profile/': temp_func,
                'list/': temp_func,
            },
        }

        collected_urls = flatten_urls(urls)
        finalized_urls = finalize_urls(collected_urls)

        expected_result = {
            'nested': {
                '<user_id>': temp_func,
                'profile': temp_func,
                'list': temp_func,
            },
            '<user_id>': temp_func,
            'profile': temp_func,
            'list': temp_func,
        }
        assert finalized_urls == expected_result

    def test_finalize_root_url(self):
        def temp_func():
            pass

        urls = {
            '/': temp_func,
        }

        collected_urls = flatten_urls(urls)
        finalized_urls = finalize_urls(collected_urls)

        expected_result = {
            '': temp_func,
        }
        assert finalized_urls == expected_result

    def test_finalize_root_url_nested(self):
        def temp_func():
            pass

        urls = {
            '/': {
                '/': {
                    'nested/<user_id>/': temp_func,
                    'nested/profile/': temp_func,
                    'nested/list/': temp_func,
                },
                '<user_id>/': temp_func,
                'profile/': temp_func,
                'list/': temp_func,
            },
        }

        collected_urls = flatten_urls(urls)
        finalized_urls = finalize_urls(collected_urls)

        expected_result = {
            'nested': {
                '<user_id>': temp_func,
                'profile': temp_func,
                'list': temp_func,
            },
            '<user_id>': temp_func,
            'profile': temp_func,
            'list': temp_func,
        }
        assert finalized_urls == expected_result

    def test_finalize_root_and_empty_url_nested(self):
        def temp_func():
            pass

        urls = {
            '/': {
                '': {
                    'nested/<user_id>/': temp_func,
                    'nested/profile/': temp_func,
                    'nested/list/': temp_func,
                },
                '/': temp_func,
                'profile/': temp_func,
                'list/': temp_func,
            },
        }

        collected_urls = flatten_urls(urls)
        finalized_urls = finalize_urls(collected_urls)

        expected_result = {
            'nested': {
                '<user_id>': temp_func,
                'profile': temp_func,
                'list': temp_func,
            },
            '': temp_func,
            'profile': temp_func,
            'list': temp_func,
        }
        assert finalized_urls == expected_result

    def test_finalize_empty_and_root_url_nested(self):
        def temp_func():
            pass

        urls = {
            '': {
                '/': {
                    'nested/<user_id>/': temp_func,
                    'nested/profile/': temp_func,
                    'nested/list/': temp_func,
                },
                '': temp_func,
                'profile/': temp_func,
                'list/': temp_func,
            },
        }

        collected_urls = flatten_urls(urls)
        finalized_urls = finalize_urls(collected_urls)

        expected_result = {
            'nested': {
                '<user_id>': temp_func,
                'profile': temp_func,
                'list': temp_func,
            },
            '': temp_func,
            'profile': temp_func,
            'list': temp_func,
        }
        assert finalized_urls == expected_result

    def test_finalize_urls(self):
        def temp_func():
            pass

        urls = {
            'user/': {
                '<user_id>/profile/<id>': temp_func,
                'profile/': temp_func,
                'list/': temp_func,
            },
            '': {
                'payments': temp_func,
                'notifications': temp_func,
            },
            'admin/v1': {
                'profile/avatar': temp_func,
                '<user_id>': temp_func,
                'users/': {
                    'list': {
                        'registered': temp_func,
                        'not-registered': temp_func,
                    },
                    'detail': {
                        'registered': temp_func,
                        'not-registered': temp_func,
                    },
                },
            },
            'admin/v2': {},
        }

        collected_urls = flatten_urls(urls)
        finalized_urls = finalize_urls(collected_urls)

        expected_result = {
            'user': {
                '<user_id>': {
                    'profile': {
                        '<id>': temp_func,
                    },
                },
                'profile': temp_func,
                'list': temp_func,
            },
            'payments': temp_func,
            'notifications': temp_func,
            'admin': {
                'v1': {
                    'profile': {
                        'avatar': temp_func,
                    },
                    '<user_id>': temp_func,
                    'users': {
                        'list': {
                            'registered': temp_func,
                            'not-registered': temp_func,
                        },
                        'detail': {
                            'registered': temp_func,
                            'not-registered': temp_func,
                        },
                    },
                },
                'v2': {},
            },
        }
        assert finalized_urls == expected_result

    def test_finalize_urls_same_pre_path_variable(self):
        def temp_func():
            pass

        urls = {
            '': temp_func,
            '<index>/': temp_func,
            '<index>/<id>/': temp_func,
        }

        collected_urls = flatten_urls(urls)
        finalized_urls = finalize_urls(collected_urls)
        expected_result = {
            '': temp_func,
            '<index>': {
                '': temp_func,
                '<id>': temp_func,
            },
        }
        assert finalized_urls == expected_result

    def test_finalize_urls_with_same_level_path_variables(self):
        def temp_func():
            pass

        urls1 = {
            'user': {
                '<index1>/': temp_func,
                '<index2>/': temp_func,
            },
        }
        urls2 = {
            'user': {
                '<index1>/': {'detail': temp_func},
                '<index2>/': temp_func,
                '<index3>/': {'detail': temp_func},
                '<index4>/': {'detail': temp_func},
            },
        }

        try:
            finalize_urls(flatten_urls(urls1))
        except PantherError as exc:
            assert exc.args[0] == (
                "URLs can't have same-level path variables that point to an endpoint: "
                '\n\t- /user/<index1>'
                '\n\t- /user/<index2>'
            )
        else:
            assert False

        try:
            finalize_urls(flatten_urls(urls2))
        except PantherError as exc:
            assert exc.args[0] == (
                "URLs can't have same-level path variables that point to a dict: "
                '\n\t- /user/<index1>'
                '\n\t- /user/<index3>'
                '\n\t- /user/<index4>'
            )
        else:
            assert False

    # Find Endpoint
    def test_find_endpoint_root_url(self):
        def temp_func():
            pass

        from panther.configs import config

        config.URLS = {
            '': temp_func,
        }
        _func, _ = find_endpoint('')

        assert _func == temp_func

    def test_find_endpoint_success(self):
        def user_id_profile_id():
            pass

        def user_profile():
            pass

        def payment():
            pass

        def admin_v1_profile_avatar():
            pass

        def admin_v1_id():
            pass

        def admin_v2_users_list_registered():
            pass

        def admin_v2_users_detail_not_registered():
            pass

        from panther.configs import config

        config.URLS = {
            'user': {
                '<user_id>': {
                    'profile': {
                        '<id>': user_id_profile_id,
                    },
                },
                'profile': user_profile,
                'list': ...,
            },
            'payments': payment,
            'notifications': ...,
            'admin': {
                'v1': {
                    'profile': {
                        'avatar': admin_v1_profile_avatar,
                    },
                    '<user_id>': admin_v1_id,
                    'users': {
                        'list': {
                            'registered': admin_v2_users_list_registered,
                            'not-registered': ...,
                        },
                        'detail': {
                            'registered': ...,
                            'not-registered': admin_v2_users_detail_not_registered,
                        },
                    },
                },
            },
        }
        user_id_profile_id_func, _ = find_endpoint(f'user/{random.randint(0, 100)}/profile/{random.randint(2, 100)}')
        user_profile_func, _ = find_endpoint('user/profile/')
        payment_func, _ = find_endpoint('payments/')
        admin_v1_profile_avatar_func, _ = find_endpoint('admin/v1/profile/avatar')
        admin_v1_id_func, _ = find_endpoint(f'admin/v1/{random.randint(0, 100)}')
        admin_v2_users_list_registered_func, _ = find_endpoint('admin/v1/users/list/registered/')
        admin_v2_users_detail_not_registered_func, _ = find_endpoint('admin/v1/users/detail/not-registered')

        assert user_id_profile_id_func == user_id_profile_id
        assert user_profile_func == user_profile
        assert payment_func == payment
        assert admin_v1_profile_avatar_func == admin_v1_profile_avatar
        assert admin_v1_id_func == admin_v1_id
        assert admin_v2_users_list_registered_func == admin_v2_users_list_registered
        assert admin_v2_users_detail_not_registered_func == admin_v2_users_detail_not_registered

    def test_find_endpoint_success_path(self):
        def user_id_profile_id():
            pass

        def user_profile():
            pass

        def payment():
            pass

        def admin_v1_profile_avatar():
            pass

        def admin_v1_id():
            pass

        def admin_v2_users_list_registered():
            pass

        def admin_v2_users_detail_not_registered():
            pass

        from panther.configs import config

        config.URLS = {
            'user': {
                '<user_id>': {
                    'profile': {
                        '<id>': user_id_profile_id,
                    },
                },
                'profile': user_profile,
                'list': ...,
            },
            'payments': payment,
            'notifications': ...,
            'admin': {
                'v1': {
                    'profile': {
                        'avatar': admin_v1_profile_avatar,
                    },
                    '<user_id>': admin_v1_id,
                    '<user_id2>': {
                        'list': {
                            '<registered2>': {},
                            '<registered1>': admin_v2_users_list_registered,
                            'registered': admin_v2_users_list_registered,
                        },
                    },
                    'users': {
                        'list': {
                            'registered': admin_v2_users_list_registered,
                            'not-registered': ...,
                        },
                        'detail': {
                            'registered': ...,
                            'not-registered': admin_v2_users_detail_not_registered,
                        },
                    },
                },
            },
        }
        _, user_id_profile_id_path = find_endpoint(f'user/{random.randint(0, 100)}/profile/{random.randint(2, 100)}')
        _, user_profile_path = find_endpoint('user/profile/')
        _, payment_path = find_endpoint('payments/')
        _, admin_v1_profile_avatar_path = find_endpoint('admin/v1/profile/avatar')
        _, admin_v1_id_path = find_endpoint(f'admin/v1/{random.randint(0, 100)}')
        _, admin_v1_id_registered_path = find_endpoint(f'admin/v1/{random.randint(0, 100)}/list/registered')
        _, admin_v1_id_registered1_path = find_endpoint(f'admin/v1/{random.randint(0, 100)}/list/1/')
        _, admin_v2_users_list_registered_path = find_endpoint('admin/v1/users/list/registered/')
        _, admin_v2_users_detail_not_registered_path = find_endpoint('admin/v1/users/detail/not-registered')

        assert user_id_profile_id_path == 'user/<user_id>/profile/<id>'
        assert user_profile_path == 'user/profile'
        assert payment_path == 'payments'
        assert admin_v1_profile_avatar_path == 'admin/v1/profile/avatar'
        assert admin_v1_id_path == 'admin/v1/<user_id>'
        assert admin_v1_id_registered_path == 'admin/v1/<user_id2>/list/registered'
        assert admin_v1_id_registered1_path == 'admin/v1/<user_id2>/list/<registered1>'
        assert admin_v2_users_list_registered_path == 'admin/v1/users/list/registered'
        assert admin_v2_users_detail_not_registered_path == 'admin/v1/users/detail/not-registered'

    def test_find_endpoint_not_found(self):
        def temp_func():
            pass

        from panther.configs import config

        config.URLS = {
            'user': {
                'list': temp_func,
            },
        }
        user_id_profile_id_func, _ = find_endpoint(f'user/{random.randint(0, 100)}/profile/{random.randint(2, 100)}')
        user_profile_func, _ = find_endpoint('user/profile/')
        payment_func, _ = find_endpoint('payments/')
        admin_v1_profile_avatar_func, _ = find_endpoint('admin/v1/profile/avatar')
        admin_v1_id_func, _ = find_endpoint(f'admin/v1/{random.randint(0, 100)}')
        admin_v2_users_list_registered_func, _ = find_endpoint('admin/v1/users/list/registered/')
        admin_v2_users_detail_not_registered_func, _ = find_endpoint('admin/v1/users/detail/not-registered')

        assert user_id_profile_id_func is None
        assert user_profile_func is None
        assert payment_func is None
        assert admin_v1_profile_avatar_func is None
        assert admin_v1_id_func is None
        assert admin_v2_users_list_registered_func is None
        assert admin_v2_users_detail_not_registered_func is None

    def test_find_endpoint_not_found_path(self):
        def temp_func():
            pass

        from panther.configs import config

        config.URLS = {
            'user': {
                'list': temp_func,
            },
        }
        _, user_id_profile_id_path = find_endpoint(f'user/{random.randint(0, 100)}/profile/{random.randint(2, 100)}')
        _, user_profile_path = find_endpoint('user/profile/')
        _, payment_path = find_endpoint('payments/')
        _, admin_v1_profile_avatar_path = find_endpoint('admin/v1/profile/avatar')
        _, admin_v1_id_path = find_endpoint(f'admin/v1/{random.randint(0, 100)}')
        _, admin_v2_users_list_registered_path = find_endpoint('admin/v1/users/list/registered/')
        _, admin_v2_users_detail_not_registered_path = find_endpoint('admin/v1/users/detail/not-registered')

        assert user_id_profile_id_path == ''
        assert user_profile_path == ''
        assert payment_path == ''
        assert admin_v1_profile_avatar_path == ''
        assert admin_v1_id_path == ''
        assert admin_v2_users_list_registered_path == ''
        assert admin_v2_users_detail_not_registered_path == ''

    def test_find_endpoint_not_found_last_is_path_variable(self):
        def temp_func():
            pass

        from panther.configs import config

        config.URLS = {
            'user': {
                '<name>': temp_func,
            },
        }
        user_id_profile_id_func, _ = find_endpoint(f'user/{random.randint(0, 100)}/profile/{random.randint(2, 100)}')
        user_profile_func, _ = find_endpoint('user/ali/')
        payment_func, _ = find_endpoint('payments/')
        admin_v1_profile_avatar_func, _ = find_endpoint('admin/v1/profile/avatar')
        admin_v1_id_func, _ = find_endpoint(f'admin/v1/{random.randint(0, 100)}')
        admin_v2_users_list_registered_func, _ = find_endpoint('admin/v1/users/list/registered/')
        admin_v2_users_detail_not_registered_func, _ = find_endpoint('admin/v1/users/detail/not-registered')

        assert user_id_profile_id_func is None
        assert user_profile_func is not None
        assert payment_func is None
        assert admin_v1_profile_avatar_func is None
        assert admin_v1_id_func is None
        assert admin_v2_users_list_registered_func is None
        assert admin_v2_users_detail_not_registered_func is None

    def test_find_endpoint_not_found_path_last_is_path_variable(self):
        def temp_func():
            pass

        from panther.configs import config

        config.URLS = {
            'user': {
                '<name>': temp_func,
            },
        }
        _, user_id_profile_id_path = find_endpoint(f'user/{random.randint(0, 100)}/profile/{random.randint(2, 100)}')
        _, user_profile_path = find_endpoint('user/ali/')
        _, payment_path = find_endpoint('payments/')
        _, admin_v1_profile_avatar_path = find_endpoint('admin/v1/profile/avatar')
        _, admin_v1_id_path = find_endpoint(f'admin/v1/{random.randint(0, 100)}')
        _, admin_v2_users_list_registered_path = find_endpoint('admin/v1/users/list/registered/')
        _, admin_v2_users_detail_not_registered_path = find_endpoint('admin/v1/users/detail/not-registered')

        assert user_id_profile_id_path == ''
        assert user_profile_path != ''
        assert payment_path == ''
        assert admin_v1_profile_avatar_path == ''
        assert admin_v1_id_path == ''
        assert admin_v2_users_list_registered_path == ''
        assert admin_v2_users_detail_not_registered_path == ''

    def test_find_endpoint_not_found_too_many(self):
        def temp_func():
            pass

        from panther.configs import config

        config.URLS = {
            'user/name': temp_func,
        }
        func, path = find_endpoint('user/name/troublemaker')

        assert path == ''
        assert func is None

    def test_find_endpoint_not_found_not_enough(self):
        def temp_func():
            pass

        from panther.configs import config

        config.URLS = {
            'user/name': temp_func,
        }
        func, path = find_endpoint('user/')

        assert path == ''
        assert func is None

    def test_find_endpoint_same_pre_path_variable(self):
        def temp_1():
            pass

        def temp_2():
            pass

        def temp_3():
            pass

        from panther.configs import config

        config.URLS = {
            '': temp_1,
            '<index>': {
                '': temp_2,
                '<id>': temp_3,
            },
        }
        temp_1_func, _ = find_endpoint('')
        temp_2_func, _ = find_endpoint(f'{random.randint(2, 100)}')
        temp_3_func, _ = find_endpoint(f'{random.randint(2, 100)}/{random.randint(2, 100)}')

        assert temp_1_func == temp_1
        assert temp_2_func == temp_2
        assert temp_3_func == temp_3

    def test_find_endpoint_same_pre_path_variable_path(self):
        def temp_1():
            pass

        def temp_2():
            pass

        def temp_3():
            pass

        from panther.configs import config

        config.URLS = {
            '': temp_1,
            '<index>': {
                '': temp_2,
                '<id>': temp_3,
            },
        }
        _, temp_1_path = find_endpoint('')
        _, temp_2_path = find_endpoint(f'{random.randint(2, 100)}')
        _, temp_3_path = find_endpoint(f'{random.randint(2, 100)}/{random.randint(2, 100)}')

        assert temp_1_path == ''
        assert temp_2_path == '<index>'
        assert temp_3_path == '<index>/<id>'

    def test_find_endpoint_same_pre_key(self):
        def temp_1():
            pass

        def temp_2():
            pass

        def temp_3():
            pass

        from panther.configs import config

        config.URLS = {
            '': temp_1,
            'hello': {
                '': temp_2,
                '<id>': temp_3,
            },
        }
        temp_1_func, _ = find_endpoint('')

        temp_2_func, _ = find_endpoint('hello')
        temp_3_func, _ = find_endpoint(f'hello/{random.randint(2, 100)}')

        assert temp_1_func == temp_1
        assert temp_2_func == temp_2
        assert temp_3_func == temp_3

    def test_find_endpoint_same_pre_key_path(self):
        def temp_1():
            pass

        def temp_2():
            pass

        def temp_3():
            pass

        from panther.configs import config

        config.URLS = {
            '': temp_1,
            'hello': {
                '': temp_2,
                '<id>': temp_3,
            },
        }
        _, temp_1_path = find_endpoint('')

        _, temp_2_path = find_endpoint('hello')
        _, temp_3_path = find_endpoint(f'hello/{random.randint(2, 100)}')

        assert temp_1_path == ''
        assert temp_2_path == 'hello'
        assert temp_3_path == 'hello/<id>'

    def test_find_endpoint_with_params(self):
        def user_id_profile_id():
            pass

        from panther.configs import config

        config.URLS = {
            'user': {
                '<user_id>': {
                    'profile': user_id_profile_id,
                },
            },
        }
        user_id_profile_id_func, _ = find_endpoint(f'user/{random.randint(0, 100)}/profile?name=ali')

        assert user_id_profile_id_func == user_id_profile_id

    # Collect PathVariables
    def test_collect_path_variables(self):
        def temp_func():
            pass

        from panther.configs import config

        config.URLS = {
            'user': {
                '<user_id>': {
                    'profile': {
                        '<id>': temp_func,
                    },
                },
            },
        }

        _user_id = random.randint(0, 100)
        _id = random.randint(0, 100)
        request_path = f'user/{_user_id}/profile/{_id}'

        _, found_path = find_endpoint(request_path)
        request = BaseRequest(scope={'path': request_path}, receive=lambda x: x, send=lambda x: x)
        request.collect_path_variables(found_path=found_path)
        path_variables = request.path_variables

        assert isinstance(path_variables, dict)

        assert 'user_id' in path_variables
        assert 'id' in path_variables

        assert path_variables['user_id'] == str(_user_id)
        assert path_variables['id'] == str(_id)

    # Complete test ready for benchmarking
    def test_find_endpoint_complete(self):
        def _():
            pass

        def _3():
            pass

        def _1():
            pass

        def _4():
            pass

        def _159():
            pass

        def _16():
            pass

        def _1710():
            pass

        def _18():
            pass

        def _19():
            pass

        def _211():
            pass

        def _2121516():
            pass

        def _2141718():
            pass

        def _220():
            pass

        def _22122():
            pass

        from panther.configs import config

        config.URLS = {
            '': _,
            '0': {'21': {}},
            '1': {
                '': _1,
                '<5>': {'<9>': _159},
                '6': _16,
                '7': {'10': _1710},
                '<8>': _18,
            },
            '<2>': {
                '11': _211,
                '12': {'15': {'16': _2121516}},
                '<14>': {'<17>': {'<18>': _2141718}},
                '19': {'': _19},
                '<20>': _220,
                '21': {'<22>': {'': _22122}},
            },
            '3': _3,
            '<4>': _4,
        }

        test_cases = {
            '': (_, ''),
            '0': ENDPOINT_NOT_FOUND,
            '0/21': ENDPOINT_NOT_FOUND,
            '3': (_3, '3'),
            '_4': (_4, '<4>'),
            '1': (_1, '1'),
            '1/_5/_9': (_159, '1/<5>/<9>'),
            '1/6': (_16, '1/6'),
            '1/7/10': (_1710, '1/7/10'),
            '1/7': ENDPOINT_NOT_FOUND,
            '1/_8': (_18, '1/<8>'),
            '_2/11': (_211, '<2>/11'),
            '_2/12/15/16': (_2121516, '<2>/12/15/16'),
            '_2/_14/_17/_18': (_2141718, '<2>/<14>/<17>/<18>'),
            '_2/19': (_19, '<2>/19'),
            '_2/_20': (_220, '<2>/<20>'),
            '_2/21/_22': (_22122, '<2>/21/<22>'),
            # # #
            '/': (_, ''),
            '/0/': ENDPOINT_NOT_FOUND,
            '/0/21/': ENDPOINT_NOT_FOUND,
            '/3/': (_3, '3'),
            '/_4/': (_4, '<4>'),
            '/1/': (_1, '1'),
            '/1/_5/_9/': (_159, '1/<5>/<9>'),
            '/1/6/': (_16, '1/6'),
            '/1/7/10/': (_1710, '1/7/10'),
            '/1/7/': ENDPOINT_NOT_FOUND,
            '/1/_8/': (_18, '1/<8>'),
            '/_2/11/': (_211, '<2>/11'),
            '/_2/12/15/16/': (_2121516, '<2>/12/15/16'),
            '/_2/_14/_17/_18/': (_2141718, '<2>/<14>/<17>/<18>'),
            '/_2/_14/_17/': ENDPOINT_NOT_FOUND,
            '/_2/19/': (_19, '<2>/19'),
            '/_2/_20/': (_220, '<2>/<20>'),
            '/_2/21/_22/': (_22122, '<2>/21/<22>'),
        }

        for test_url, expected in test_cases.items():
            actual = find_endpoint(path=test_url)
            assert actual == expected



import typing
from collections.abc import Callable
from dataclasses import dataclass, field
from datetime import timedelta
from pathlib import Path

import jinja2
from pydantic import BaseModel as PydanticBaseModel


class JWTConfig:
    def __init__(
        self,
        key: str,
        algorithm: str = 'HS256',
        life_time: timedelta | int = timedelta(days=1),
        refresh_life_time: timedelta | int | None = None,
    ):
        self.key = key
        self.algorithm = algorithm
        self.life_time = int(life_time.total_seconds()) if isinstance(life_time, timedelta) else life_time

        if refresh_life_time:
            if isinstance(refresh_life_time, timedelta):
                self.refresh_life_time = refresh_life_time.total_seconds()
            else:
                self.refresh_life_time = refresh_life_time
        else:
            self.refresh_life_time = self.life_time * 2

    def __eq__(self, other):
        return bool(
            self.key == other.key
            and self.algorithm == other.algorithm
            and self.life_time == other.life_time
            and self.refresh_life_time == other.refresh_life_time,
        )


class QueryObservable:
    observers = []

    @classmethod
    def observe(cls, observer):
        cls.observers.append(observer)

    @classmethod
    def update(cls):
        for observer in cls.observers:
            observer._reload_bases(parent=config.QUERY_ENGINE)


@dataclass
class Config:
    BASE_DIR: Path = Path()
    MONITORING: bool = False
    LOG_QUERIES: bool = False
    THROTTLING = None  # type: panther.throttling.Throttle
    SECRET_KEY: str | None = None
    HTTP_MIDDLEWARES: list = field(default_factory=list)
    WS_MIDDLEWARES: list = field(default_factory=list)
    USER_MODEL: type[PydanticBaseModel] | None = None
    AUTHENTICATION: type[PydanticBaseModel] | None = None
    WS_AUTHENTICATION: type[PydanticBaseModel] | None = None
    JWT_CONFIG: JWTConfig | None = None
    MODELS: list = field(default_factory=list)
    FLAT_URLS: dict = field(default_factory=dict)
    URLS: dict = field(default_factory=dict)
    WEBSOCKET_CONNECTIONS: Callable | None = None
    BACKGROUND_TASKS: bool = False
    HAS_WS: bool = False
    TIMEZONE: str = 'UTC'
    TEMPLATES_DIR: str | list[str] = '.'
    JINJA_ENVIRONMENT: jinja2.Environment | None = None
    AUTO_REFORMAT: bool = False
    QUERY_ENGINE: Callable | None = None
    DATABASE: Callable | None = None

    def refresh(self):
        """
        Reset built-in fields and remove any custom (non-built-in) attributes.
        * In some tests we need to `refresh` the `config` values
        """
        builtin_fields = set(self.__dataclass_fields__)
        current_fields = set(self.__dict__)

        # Reset built-in fields
        for field_name in builtin_fields:
            field_def = self.__dataclass_fields__[field_name]
            default = field_def.default_factory() if callable(field_def.default_factory) else field_def.default
            setattr(self, field_name, default)

        # Delete custom attributes
        for field_name in current_fields - builtin_fields:
            delattr(self, field_name)

    def vars(self) -> dict[str, typing.Any]:
        """Return all config variables (built-in + custom)."""
        return dict(self.__dict__)

    def __setattr__(self, key, value):
        super().__setattr__(key, value)
        if key == 'QUERY_ENGINE' and value:
            QueryObservable.update()

    def __getattr__(self, item: str):
        try:
            return object.__getattribute__(self, item)
        except AttributeError:
            return None

    def __setitem__(self, key, value):
        setattr(self, key.upper(), value)

    def __getitem__(self, item):
        return getattr(self, item.upper())


config = Config()



from unittest import IsolatedAsyncioTestCase

from panther import Panther
from panther.app import API, GenericAPI
from panther.configs import config
from panther.exceptions import PantherError
from panther.permissions import BasePermission
from panther.request import Request
from panther.test import APIClient


class AlwaysDeniedPermission(BasePermission):
    async def __call__(self, request: Request) -> bool:
        return False


class NotInheritedPermission:
    async def __call__(self, request: Request) -> bool:
        return False


class SyncPermission(BasePermission):
    def __call__(self, request: Request) -> bool:
        return False


class WithoutParamPermission:
    def __call__(self) -> bool:
        return False


def without_param_permission():
    return False


class WithoutCallPermission:
    pass


def sync_permission(req):
    return False


async def accept_permission(req):
    return True


async def deny_permission(req):
    return True


@API()
async def without_permission_api(request: Request):
    return request.user


@API(permissions=[AlwaysDeniedPermission])
async def denied_permission_api(request: Request):
    return request.user


@API(permissions=AlwaysDeniedPermission)
async def single_denied_permission_api(request: Request):
    return request.user


@API(permissions=[NotInheritedPermission])
async def not_inherited_permission_api(request: Request):
    return request.user


@API(permissions=[accept_permission, accept_permission, AlwaysDeniedPermission])
async def multiple_permissions_api(request: Request):
    return request.user


urls = {
    'without': without_permission_api,
    'denied-permission': denied_permission_api,
    'single-denied-permission': single_denied_permission_api,
    'not-inherited-permission': not_inherited_permission_api,
    'multiple-permission': multiple_permissions_api,
}


class TestPermission(IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        app = Panther(__name__, configs=__name__, urls=urls)
        self.client = APIClient(app=app)

    @classmethod
    def tearDownClass(cls):
        config.refresh()

    async def test_without_permission(self):
        res = await self.client.get('without')
        assert res.status_code == 200
        assert res.data is None

    async def test_denied_permission(self):
        with self.assertNoLogs(level='ERROR'):
            res = await self.client.get('denied-permission')
        assert res.status_code == 403
        assert res.data['detail'] == 'Permission Denied'

    async def test_single_denied_permission(self):
        with self.assertNoLogs(level='ERROR'):
            res = await self.client.get('single-denied-permission')
        assert res.status_code == 403
        assert res.data['detail'] == 'Permission Denied'

    async def test_not_inherited_permission(self):
        with self.assertNoLogs(level='ERROR'):
            res = await self.client.get('not-inherited-permission')
        assert res.status_code == 403
        assert res.data['detail'] == 'Permission Denied'

    async def test_multiple_permission(self):
        with self.assertNoLogs(level='ERROR'):
            res = await self.client.get('multiple-permission')
        assert res.status_code == 403
        assert res.data['detail'] == 'Permission Denied'

    async def test_sync_permission(self):
        try:

            @API(permissions=[SyncPermission])
            async def sync_permission_api(request: Request):
                return request.user
        except PantherError as e:
            assert e.args[0] == 'SyncPermission.__call__() should be `async`'
        else:
            assert False

    async def test_sync_permission_classbased(self):
        try:

            class SyncPermissionAPI(GenericAPI):
                permissions = [SyncPermission]
        except PantherError as e:
            assert e.args[0] == 'SyncPermission.__call__() should be `async`'
        else:
            assert False

    async def test_without_call_permission(self):
        try:

            @API(permissions=[WithoutCallPermission])
            async def permission_api(request: Request):
                return request.user
        except PantherError as e:
            assert e.args[0] == 'WithoutCallPermission must implement __call__() method.'
        else:
            assert False

    async def test_without_call_permission_classbased(self):
        try:

            class PermissionAPI(GenericAPI):
                permissions = [WithoutCallPermission]
        except PantherError as e:
            assert e.args[0] == 'WithoutCallPermission must implement __call__() method.'
        else:
            assert False

    async def test_sync_function_permission(self):
        try:

            @API(permissions=[sync_permission])
            async def permission_api(request: Request):
                return request.user
        except PantherError as e:
            assert e.args[0] == 'sync_permission() should be `async`'
        else:
            assert False

    async def test_sync_function_permission_classbased(self):
        try:

            class PermissionAPI(GenericAPI):
                permissions = [sync_permission]
        except PantherError as e:
            assert e.args[0] == 'sync_permission() should be `async`'
        else:
            assert False

    async def test_class_without_param_permission(self):
        try:

            @API(permissions=[WithoutParamPermission])
            async def permission_api(request: Request):
                return request.user
        except PantherError as e:
            assert e.args[0] == 'WithoutParamPermission.__call__() requires 2 positional argument(s) (self, request).'
        else:
            assert False

    async def test_class_without_param_permission_classbased(self):
        try:

            class PermissionAPI(GenericAPI):
                permissions = [WithoutParamPermission]
        except PantherError as e:
            assert e.args[0] == 'WithoutParamPermission.__call__() requires 2 positional argument(s) (self, request).'
        else:
            assert False

    async def test_function_without_param_permission(self):
        try:

            @API(permissions=[without_param_permission])
            async def permission_api(request: Request):
                return request.user
        except PantherError as e:
            assert e.args[0] == 'without_param_permission() requires 1 positional argument(s) (request).'
        else:
            assert False

    async def test_function_without_param_permission_classbased(self):
        try:

            class PermissionAPI(GenericAPI):
                permissions = [without_param_permission]
        except PantherError as e:
            assert e.args[0] == 'without_param_permission() requires 1 positional argument(s) (request).'
        else:
            assert False



import logging
import platform

from rich import print as rprint

from panther.configs import Config
from panther.exceptions import PantherError

logger = logging.getLogger('panther')

if platform.system() == 'Windows':
    h = '|'
    v = '_'
    tr = ' '
    tl = ' '
    br = ' '
    bl = ' '
else:
    h = '│'
    v = '─'
    tr = '╮'
    tl = '╭'
    br = '╯'
    bl = '╰'

top = f'{tl}{60 * v}{tr}'
bottom = f'{bl}{60 * v}{br}'

logo = rf"""{top}
{h}     ____                 __    __                          {h}
{h}    /\  _`\              /\ \__/\ \                         {h}
{h}    \ \ \L\ \ __      ___\ \ ,_\ \ \___      __   _ __      {h}
{h}     \ \ ,__/'__`\  /' _ `\ \ \/\ \  _ `\  /'__`\/\`'__\    {h}
{h}      \ \ \/\ \L\.\_/\ \/\ \ \ \_\ \ \ \ \/\  __/\ \ \/     {h}
{h}       \ \_\ \__/.\_\ \_\ \_\ \__\\ \_\ \_\ \____\\ \_\     {h}
{h}        \/_/\/__/\/_/\/_/\/_/\/__/ \/_/\/_/\/____/ \/_/     {h}
{h}                                                            {h}"""

help_message = rf"""{logo}
{h}   Usage: panther <command> \[options]                       {h}
{h}                                                            {h}
{h}   Commands:                                                {h}
{h}       - create \[project_name] \[directory]                  {h}
{h}           Create a new Panther project.                    {h}
{h}           * Interactive mode if no arguments provided.     {h}
{h}           * Non-interactive if project_name and directory  {h}
{h}             are specified (default directory: .).          {h}
{h}           Example:                                         {h}
{h}               - `panther create`                           {h}
{h}               - `panther create myapp myapp`               {h}
{h}                                                            {h}
{h}       - run <app> \[options]                                {h}
{h}           Run your Panther project using Uvicorn.          {h}
{h}           * app: address of your application.              {h}
{h}           * options: Check `uvicorn --help` for options.   {h}
{h}           * `panther run` is alias of `uvicorn`.           {h}
{h}           Example: `panther run main:app --reload`         {h}
{h}                                                            {h}
{h}       - shell <application_file>                           {h}       
{h}           Start an interactive Python shell with your app. {h}
{h}           * application_file: path to your main app file.  {h}           
{h}           Example: `panther shell main.py`                 {h}
{h}                                                            {h}
{h}       - createuser <application_file>                      {h}       
{h}           Create a new user based on USER_MODEl from your, {h}
{h}           configs. (default is panther.db.models.BaseUser) {h}
{h}           * application_file: path to your main app file.  {h}           
{h}           Example: `panther createuser main.py`            {h}
{h}                                                            {h}
{h}       - monitor                                            {h}
{h}           Display real-time request monitoring.            {h}
{h}                                                            {h}
{h}       - version | --version                                {h}
{h}           Display the current version of Panther.          {h}
{h}                                                            {h}
{h}       - help | h | --help | -h                             {h}
{h}           Show this help message and exit.                 {h}
{bottom}
"""


def import_error(message: str | Exception, package: str | None = None) -> PantherError:
    msg = str(message)
    if package:
        msg += f' -> Hint: `pip install {package}`'
    return PantherError(msg)


def cli_error(message: str | Exception) -> None:
    logger.error(message)
    logger.info('Use "panther -h" for more help')


def cli_warning(message: str | Exception, hint: str = None) -> None:
    logger.warning(message)
    if hint:
        logger.info(hint)


def cli_info(message: str) -> None:
    logger.info(message)
    logger.info('Use "panther -h" for more help\n')


def print_help_message():
    rprint(help_message)


def print_info(config: Config):
    from panther.db.connections import redis

    mo = config.MONITORING
    lq = config.LOG_QUERIES
    bt = config.BACKGROUND_TASKS
    ws = config.HAS_WS
    rd = redis.is_connected
    bd = f'{config.BASE_DIR!s:<41}'
    if len(bd) > 41:
        bd = f'{bd[:38]}...'

    # Monitoring
    monitor = f'{h} * Run "panther monitor" in another session for Monitoring  {h}\n' if config.MONITORING else None

    # Uvloop
    uvloop_msg = None
    if platform.system() != 'Windows':
        try:
            import uvloop
        except ImportError:
            uvloop_msg = (
                f'{h} * You may want to install `uvloop` for better performance  {h}\n'
                f'{h}   `pip install uvloop`                                     {h}\n'
            )

    # Gunicorn if Websocket
    gunicorn_msg = None
    if config.HAS_WS:
        try:
            import gunicorn

            gunicorn_msg = f'{h} * You have WS, so make sure to run gunicorn with --preload  {h}\n'
        except ImportError:
            pass

    # Message
    info_message = f"""{logo}
{h}   Redis: {rd}                                       \t     {h}
{h}   Websocket: {ws}                                   \t     {h}
{h}   Monitoring: {mo}                                  \t     {h}
{h}   Log Queries: {lq}                                 \t     {h}
{h}   Background Tasks: {bt}                            \t     {h}
{h}   Base directory: {bd}{h}
"""
    if monitor:
        info_message += monitor
    if uvloop_msg:
        info_message += uvloop_msg
    if gunicorn_msg:
        info_message += gunicorn_msg

    info_message += bottom
    rprint(info_message)



import platform
import tempfile
import unittest
from pathlib import Path

from panther._utils import detect_mime_type
from panther.file_handler import File


class TestFile(unittest.TestCase):
    def test_file_init_with_bytes(self):
        data = b'hello world'
        f = File(file_name='test.txt', content_type='text/plain', file=data)
        assert f.file_name == 'test.txt'
        assert f.content_type == 'text/plain'
        assert f.file == data
        assert f.size == len(data)

    def test_file_init_with_path(self):
        with tempfile.TemporaryDirectory() as tmp_dir_name:
            file_path = Path(tmp_dir_name) / 'foo.txt'
            file_path.write_bytes(b'abc123')
            f = File(file_name=str(file_path), content_type='text/plain')
            assert f.file_name == str(file_path)
            assert f.content_type == 'text/plain'
            assert f.size == 6

    def test_file_read_and_seek(self):
        with tempfile.TemporaryDirectory() as tmp_dir_name:
            file_path = Path(tmp_dir_name) / 'bar.txt'
            file_path.write_bytes(b'abcdef')
            f = File(file_name=str(file_path), content_type='text/plain')
            with f:
                assert f.read(3) == b'abc'
                assert f.tell() == 3
                f.seek(0)
                assert f.read() == b'abcdef'

    def test_file_write_in_memory(self):
        f = File(file_name='baz.txt', content_type='text/plain', file=b'foo')
        f.write(b'bar')
        assert f.file == b'foobar'
        assert f.size == 6

    def test_file_write_on_disk_raises(self):
        with tempfile.TemporaryDirectory() as tmp_dir_name:
            file_path = Path(tmp_dir_name) / 'baz.txt'
            file_path.write_bytes(b'foo')
            f = File(file_name=str(file_path), content_type='text/plain')
            with f:
                try:
                    f.write(b'bar')
                    assert False, 'Expected IOError to be raised'
                except IOError:
                    pass

    def test_file_save_overwrite(self):
        with tempfile.TemporaryDirectory() as tmp_dir_name:
            f = File(file_name='test.txt', content_type='text/plain', file=b'data')
            save_path = Path(tmp_dir_name) / 'out.txt'
            result_path = f.save(str(save_path), overwrite=True)
            assert Path(result_path).read_bytes() == b'data'
            # Overwrite again
            f2 = File(file_name='test.txt', content_type='text/plain', file=b'new')
            result_path2 = f2.save(str(save_path), overwrite=True)
            assert Path(result_path2).read_bytes() == b'new'
            assert result_path == result_path2

    def test_file_save_no_overwrite(self):
        with tempfile.TemporaryDirectory() as tmp_dir_name:
            f = File(file_name='test.txt', content_type='text/plain', file=b'data')
            save_path = Path(tmp_dir_name) / 'out.txt'
            result_path1 = f.save(str(save_path), overwrite=False)
            assert Path(result_path1).read_bytes() == b'data'
            # Save again, should create a new file with timestamp
            f2 = File(file_name='test.txt', content_type='text/plain', file=b'new')
            result_path2 = f2.save(str(save_path), overwrite=False)
            assert Path(result_path2).read_bytes() == b'new'
            assert result_path1 != result_path2

    def test_file_save_idempotent(self):
        """Test that save() method is idempotent - returns same path on multiple calls"""
        with tempfile.TemporaryDirectory() as tmp_dir_name:
            f = File(file_name='test.txt', content_type='text/plain', file=b'data')
            save_path = Path(tmp_dir_name) / 'out.txt'

            # First call
            result_path1 = f.save(str(save_path), overwrite=True)
            assert Path(result_path1).read_bytes() == b'data'

            # Second call - should return same path
            result_path2 = f.save(str(save_path), overwrite=True)
            assert result_path1 == result_path2

            # Third call - should return same path
            result_path3 = f.save(str(save_path), overwrite=True)
            assert result_path1 == result_path3

            # Verify file content is still the same
            assert Path(result_path1).read_bytes() == b'data'

    def test_file_save_idempotent_with_different_paths(self):
        """Test that idempotency works even when called with different paths"""
        with tempfile.TemporaryDirectory() as tmp_dir_name:
            f = File(file_name='test.txt', content_type='text/plain', file=b'data')

            # First call
            result_path1 = f.save(Path(tmp_dir_name) / 'first.txt', overwrite=True)

            # Second call with different path - should still return first path
            result_path2 = f.save(Path(tmp_dir_name) / 'second.txt', overwrite=True)
            assert result_path1 == result_path2

            # Third call with no path - should still return first path
            result_path3 = f.save(overwrite=True)
            assert result_path1 == result_path3

    def test_file_save_directory_path(self):
        """Test saving to directory path (ending with /)"""
        with tempfile.TemporaryDirectory() as tmp_dir_name:
            f = File(file_name='document.pdf', content_type='application/pdf', file=b'pdf content')

            # Save to directory
            result_path = f.save(f'{tmp_dir_name}/uploads/')
            expected_path = Path(tmp_dir_name) / 'uploads' / 'document.pdf'

            assert result_path == str(expected_path)
            assert expected_path.exists()
            assert expected_path.read_bytes() == b'pdf content'

    def test_file_save_nested_directory_path(self):
        """Test saving to nested directory path"""
        with tempfile.TemporaryDirectory() as tmp_dir_name:
            f = File(file_name='image.jpg', content_type='image/jpeg', file=b'jpeg content')

            # Save to nested directory
            result_path = f.save(f'{tmp_dir_name}/uploads/images/')
            expected_path = Path(tmp_dir_name) / 'uploads' / 'images' / 'image.jpg'

            assert result_path == str(expected_path)
            assert expected_path.exists()
            assert expected_path.read_bytes() == b'jpeg content'

    def test_file_save_directory_path_with_custom_filename(self):
        """Test saving to directory with custom filename"""
        with tempfile.TemporaryDirectory() as tmp_dir_name:
            f = File(file_name='original.txt', content_type='text/plain', file=b'content')

            # Save to directory with custom filename
            result_path = f.save(f'{tmp_dir_name}/uploads/custom_name.txt')
            expected_path = Path(tmp_dir_name) / 'uploads' / 'custom_name.txt'

            assert result_path == str(expected_path)
            assert expected_path.exists()
            assert expected_path.read_bytes() == b'content'

    def test_file_save_directory_creation(self):
        """Test that directories are created automatically"""
        with tempfile.TemporaryDirectory() as tmp_dir_name:
            f = File(file_name='test.txt', content_type='text/plain', file=b'data')

            # Try to save to non-existent directory
            target_dir = Path(tmp_dir_name) / 'new' / 'nested' / 'directory'
            result_path = f.save(str(target_dir / 'test.txt'), overwrite=True)

            # Directory should be created
            assert target_dir.exists()
            assert Path(result_path).exists()

    def test_file_save_directory_path_idempotent(self):
        """Test that directory path saving is also idempotent"""
        with tempfile.TemporaryDirectory() as tmp_dir_name:
            f = File(file_name='test.txt', content_type='text/plain', file=b'data')

            # First call
            result_path1 = f.save(f'{tmp_dir_name}/uploads/')

            # Second call - should return same path
            result_path2 = f.save(f'{tmp_dir_name}/uploads/')
            assert result_path1 == result_path2

            # Third call with different directory - should still return first path
            result_path3 = f.save(f'{tmp_dir_name}/different/')
            assert result_path1 == result_path3

    def test_file_save_without_path_uses_filename(self):
        """Test that save() without path uses the original filename"""
        with tempfile.TemporaryDirectory() as tmp_dir_name:
            original_cwd = Path.cwd()
            try:
                # Change to temp directory for testing
                import os

                os.chdir(tmp_dir_name)

                f = File(file_name='test.txt', content_type='text/plain', file=b'data')
                result_path = f.save()

                # Should save as 'test.txt' in current directory
                expected_path = Path(tmp_dir_name) / 'test.txt'
                assert Path(result_path).resolve() == expected_path.resolve()
                assert Path(result_path).exists()

            finally:
                os.chdir(original_cwd)

    def test_file_repr_and_str(self):
        f = File(file_name='foo.txt', content_type='text/plain', file=b'bar')
        s = str(f)
        assert 'foo.txt' in s
        assert 'text/plain' in s

    def test_file_ensure_buffer_error(self):
        f = File(file_name='foo.txt', content_type='text/plain')
        f.file = None
        f._file_path = None
        f._buffer = None
        try:
            f._ensure_buffer()
            assert False, 'Expected ValueError to be raised'
        except ValueError:
            pass


class TestDetectMimeType(unittest.TestCase):
    def test_extension_based_detection(self):
        with tempfile.TemporaryDirectory() as tmp_dir_name:
            file_path = Path(tmp_dir_name) / 'foo.txt'
            file_path.write_text('hello')
            assert detect_mime_type(str(file_path)) == 'text/plain'

    def test_png_magic_number(self):
        with tempfile.TemporaryDirectory() as tmp_dir_name:
            file_path = Path(tmp_dir_name) / 'img.png'
            file_path.write_bytes(b'\x89PNG\r\n\x1a\n' + b'rest')
            assert detect_mime_type(str(file_path)) == 'image/png'

    def test_pdf_magic_number(self):
        with tempfile.TemporaryDirectory() as tmp_dir_name:
            file_path = Path(tmp_dir_name) / 'doc.pdf'
            file_path.write_bytes(b'%PDF-1.4 some pdf data')
            assert detect_mime_type(str(file_path)) == 'application/pdf'

    def test_zip_magic_number(self):
        with tempfile.TemporaryDirectory() as tmp_dir_name:
            file_path = Path(tmp_dir_name) / 'archive.zip'
            file_path.write_bytes(b'PK\x03\x04rest')
            if platform.system() == 'Windows':
                assert detect_mime_type(str(file_path)) == 'application/x-zip-compressed'
            else:
                assert detect_mime_type(str(file_path)) == 'application/zip'

    def test_jpeg_magic_number(self):
        with tempfile.TemporaryDirectory() as tmp_dir_name:
            file_path = Path(tmp_dir_name) / 'img.jpg'
            file_path.write_bytes(b'\xff\xd8\xffrest')
            assert detect_mime_type(str(file_path)) == 'image/jpeg'

    def test_gif_magic_number(self):
        with tempfile.TemporaryDirectory() as tmp_dir_name:
            file_path = Path(tmp_dir_name) / 'img.gif'
            file_path.write_bytes(b'GIF89a rest')
            assert detect_mime_type(str(file_path)) == 'image/gif'

    def test_bmp_magic_number(self):
        with tempfile.TemporaryDirectory() as tmp_dir_name:
            file_path = Path(tmp_dir_name) / 'img.bmp'
            file_path.write_bytes(b'BMrest')
            assert detect_mime_type(str(file_path)) == 'image/bmp'

    def test_ico_magic_number(self):
        with tempfile.TemporaryDirectory() as tmp_dir_name:
            file_path = Path(tmp_dir_name) / 'icon'
            file_path.write_bytes(b'\x00\x00\x01\x00rest')
            assert detect_mime_type(str(file_path)) == 'image/x-icon'

    def test_tiff_magic_number(self):
        with tempfile.TemporaryDirectory() as tmp_dir_name:
            file_path = Path(tmp_dir_name) / 'img.tiff'
            file_path.write_bytes(b'II*\x00rest')
            assert detect_mime_type(str(file_path)) == 'image/tiff'

    def test_mp4_magic_number(self):
        with tempfile.TemporaryDirectory() as tmp_dir_name:
            file_path = Path(tmp_dir_name) / 'video.mp4'
            file_path.write_bytes(b'\x00\x00\x00\x18ftyprest')
            assert detect_mime_type(str(file_path)) == 'video/mp4'

    def test_fallback_octet_stream(self):
        with tempfile.TemporaryDirectory() as tmp_dir_name:
            file_path = Path(tmp_dir_name) / 'unknown.bin'
            file_path.write_bytes(b'randomdata123456')
            assert detect_mime_type(str(file_path)) == 'application/octet-stream'



import pydantic

from panther.serializer import ModelSerializer


class OutputSchema:
    """
    Configuration class for defining API endpoint output schemas.

    This class allows you to specify the response model, status code, and other
    metadata for API endpoints to generate proper OpenAPI documentation.

    Attributes:
        model: The Pydantic model or ModelSerializer class for the response
        status_code: HTTP status code for the response
        exclude_in_docs: Whether to exclude this endpoint from OpenAPI docs (defaults to False)
        tags: List of tags for grouping endpoints in documentation (defaults to Function Name/ Class Name)
        deprecated: Whether this endpoint is marked as deprecated (defaults to False)
    """

    def __init__(
        self,
        model: type[ModelSerializer] | type[pydantic.BaseModel] | None = None,
        status_code: int | None = None,
        exclude_in_docs: bool = False,
        tags: str | None = None,
        deprecated: bool = False,
    ):
        self.model = model
        self.status_code = status_code
        self.exclude_in_docs = exclude_in_docs
        self.tags = tags
        self.deprecated = deprecated



import asyncio
from datetime import datetime, timedelta
from unittest import IsolatedAsyncioTestCase

from panther import Panther
from panther.app import API
from panther.configs import config
from panther.test import APIClient
from panther.throttling import Throttle
from panther.utils import round_datetime


@API()
async def without_throttling_api():
    return 'ok'


@API(throttling=Throttle(rate=3, duration=timedelta(seconds=1)))
async def with_throttling_api():
    return 'ok'


@API(throttling=Throttle(rate=1, duration=timedelta(seconds=1)))
async def throttling_headers_api():
    return 'ok'


THROTTLING = Throttle(rate=1, duration=timedelta(seconds=10))

urls = {
    'without-throttling': without_throttling_api,
    'with-throttling': with_throttling_api,
    'throttling-headers': throttling_headers_api,
}


class TestThrottling(IsolatedAsyncioTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        app = Panther(__name__, configs=__name__, urls=urls)
        cls.client = APIClient(app=app)

    @classmethod
    def tearDownClass(cls):
        config.refresh()

    async def test_without_throttling(self):
        throttling = config.THROTTLING
        config.THROTTLING = None  # Disable Global Throttling

        res1 = await self.client.get('without-throttling')
        assert res1.status_code == 200

        res2 = await self.client.get('without-throttling')
        assert res2.status_code == 200

        res3 = await self.client.get('without-throttling')
        assert res3.status_code == 200
        config.THROTTLING = throttling

    async def test_with_throttling(self):
        res1 = await self.client.get('with-throttling')
        assert res1.status_code == 200

        res2 = await self.client.get('with-throttling')
        assert res2.status_code == 200

        res3 = await self.client.get('with-throttling')
        assert res3.status_code == 200

        res4 = await self.client.get('with-throttling')
        assert res4.status_code == 429

        res5 = await self.client.get('with-throttling')
        assert res5.status_code == 429

        await asyncio.sleep(1)  # Sleep and try again

        res6 = await self.client.get('with-throttling')
        assert res6.status_code == 200

        res7 = await self.client.get('with-throttling')
        assert res7.status_code == 200

        res8 = await self.client.get('with-throttling')
        assert res8.status_code == 200

        res9 = await self.client.get('with-throttling')
        assert res9.status_code == 429

        res10 = await self.client.get('with-throttling')
        assert res10.status_code == 429

    async def test_throttling_header(self):
        await self.client.get('throttling-headers')

        res = await self.client.get('throttling-headers')
        assert res.status_code == 429
        reset_time = round_datetime(datetime.now(), timedelta(seconds=1)) + timedelta(seconds=1)
        assert res.headers == {
            'Content-Type': 'application/json',
            'Content-Length': '29',
            'Retry-After': str(int((reset_time - datetime.now()).total_seconds())),
            'X-RateLimit-Reset': str(int(reset_time.timestamp())),
        }

    async def test_global_throttling(self):
        res1 = await self.client.get('without-throttling')
        assert res1.status_code == 200

        res2 = await self.client.get('without-throttling')
        assert res2.status_code == 429



import asyncio
import logging
from collections.abc import AsyncGenerator, Generator
from dataclasses import dataclass
from http import cookies
from sys import version_info
from types import NoneType
from typing import Any, Literal

import jinja2

from panther.exceptions import APIError

if version_info >= (3, 11):
    from typing import LiteralString
else:
    from typing import TypeVar

    LiteralString = TypeVar('LiteralString')

import orjson as json
from pantherdb import Cursor as PantherDBCursor
from pydantic import BaseModel

from panther import status
from panther._utils import to_async_generator
from panther.configs import config
from panther.db.cursor import Cursor
from panther.pagination import Pagination

ResponseDataTypes = (
    list | tuple | set | Cursor | PantherDBCursor | dict | int | float | str | bool | bytes | NoneType | type[BaseModel]
)
IterableDataTypes = list | tuple | set | Cursor | PantherDBCursor
StreamingDataTypes = Generator | AsyncGenerator

logger = logging.getLogger('panther')


@dataclass(slots=True)
class Cookie:
    """
    path: [Optional] Indicates the path that must exist in the requested URL for the browser to send the Cookie header.
        Default is `/`
    domain: [Optional] Defines the host to which the cookie will be sent.
        Default is the host of the current document URL, not including subdomains.
    max_age: [Optional] Indicates the number of seconds until the cookie expires.
        A zero or negative number will expire the cookie immediately.
    secure: [Optional] Indicates that the cookie is sent to the server
        only when a request is made with the https: scheme (except on localhost)
    httponly: [Optional] Forbids JavaScript from accessing the cookie,
        for example, through the `Document.cookie` property.
    samesite: [Optional] Controls whether a cookie is sent with cross-site requests or not,
        `lax` is the default behavior if not specified.
    expires: [Deprecated] In HTTP version 1.1, `expires` was deprecated and replaced with the easier-to-use `max-age`
    """

    key: str
    value: str
    domain: str = None
    max_age: int = None
    secure: bool = False
    httponly: bool = False
    samesite: Literal['none', 'lax', 'strict'] = 'lax'
    path: str = '/'


class Response:
    """
    Usage Example:
        from panther.response import Response

        def my_api():
            data = {"message": "Hello, World!"}
            return Response(data=data)
    """

    content_type = 'application/json'

    def __init__(
        self,
        data: ResponseDataTypes = None,
        status_code: int = status.HTTP_200_OK,
        headers: dict | None = None,
        pagination: Pagination | None = None,
        set_cookies: Cookie | list[Cookie] | None = None,
    ):
        """
        :param data: should be an instance of ResponseDataTypes
        :param status_code: should be int
        :param headers: should be dict of headers
        :param pagination: an instance of Pagination or None
            The `pagination.template()` method will be used
        :param set_cookies: single cookie or list of cookies you want to set on the client.
            Set the `max-age` to `0` if you want to delete a cookie.
        """
        if isinstance(data, (Cursor, PantherDBCursor)):
            data = list(data)
        self.data = data
        self.status_code = status_code
        self.headers = {'Content-Type': self.content_type} | (headers or {})
        self.pagination: Pagination | None = pagination
        self.cookies = None
        if set_cookies:
            c = cookies.SimpleCookie()
            if not isinstance(set_cookies, list):
                set_cookies = [set_cookies]
            for cookie in set_cookies:
                c[cookie.key] = cookie.value
                c[cookie.key]['path'] = cookie.path
                c[cookie.key]['secure'] = cookie.secure
                c[cookie.key]['httponly'] = cookie.httponly
                c[cookie.key]['samesite'] = cookie.samesite
                if cookie.domain is not None:
                    c[cookie.key]['domain'] = cookie.domain
                if cookie.max_age is not None:
                    c[cookie.key]['max-age'] = cookie.max_age
            self.cookies = [(b'Set-Cookie', cookie.OutputString().encode()) for cookie in c.values()]

    def __str__(self):
        if len(data := str(self.data)) > 30:
            data = f'{data:.27}...'
        return f'Response(status_code={self.status_code}, data={data})'

    __repr__ = __str__

    @property
    def body(self) -> bytes:
        def default(obj: Any):
            if isinstance(obj, BaseModel):
                return obj.model_dump()
            if isinstance(obj, (Cursor, PantherDBCursor)):
                return list(obj)
            if isinstance(obj, bytes):
                return f'raw bytes is not JSON serializable ({len(obj)} bytes)'
            raise TypeError(f'Type {type(obj)} not serializable')

        if isinstance(self.data, bytes):
            return self.data
        if self.data is None:
            return b''
        return json.dumps(self.data, default=default)

    @property
    def bytes_headers(self) -> list[tuple[bytes, bytes]]:
        headers = {'Content-Length': len(self.body)} | self.headers
        result = [(k.encode(), str(v).encode()) for k, v in headers.items()]
        if self.cookies:
            result += self.cookies
        return result

    async def send(self, send, receive):
        await send({'type': 'http.response.start', 'status': self.status_code, 'headers': self.bytes_headers})
        await send({'type': 'http.response.body', 'body': self.body, 'more_body': False})

    async def serialize_output(self, output_model: type[BaseModel]):
        """Serializes response data using the given output_model."""

        async def handle_output(obj):
            output = output_model(**obj) if isinstance(obj, dict) else output_model(**obj.model_dump())
            if hasattr(output_model, 'to_response'):
                return await output.to_response(instance=obj, data=output.model_dump())
            return output.model_dump()

        if isinstance(self.data, dict) or isinstance(self.data, BaseModel):
            self.data = await handle_output(self.data)

        elif isinstance(self.data, IterableDataTypes):
            results = []
            for d in self.data:
                if isinstance(d, dict) or isinstance(d, BaseModel):
                    results.append(await handle_output(d))
                else:
                    msg = 'Type of Response data is not match with `output_model`.\n*hint: You may want to remove `output_model`'
                    raise TypeError(msg)
            self.data = results

        else:
            msg = (
                'Type of Response data is not match with `output_model`.\n*hint: You may want to remove `output_model`'
            )
            raise TypeError(msg)


class StreamingResponse(Response):
    """
    Usage Example:
        from panther.response import StreamingResponse
        import time

        def my_generator():
            for i in range(5):
                time.sleep(1)
                yield f"Chunk {i}\n"

        def my_api():
            return StreamingResponse(data=my_generator())
    """

    content_type = 'application/octet-stream'

    def __init__(self, *args, **kwargs):
        self.connection_closed = False
        super().__init__(*args, **kwargs)

    async def listen_to_disconnection(self, receive):
        message = await receive()
        if message['type'] == 'http.disconnect':
            self.connection_closed = True

    @property
    def bytes_headers(self) -> list[tuple[bytes, bytes]]:
        result = [(k.encode(), str(v).encode()) for k, v in self.headers.items()]
        if self.cookies:
            result += self.cookies
        return result

    @property
    async def body(self) -> AsyncGenerator:
        if not isinstance(self.data, (Generator, AsyncGenerator)):
            raise TypeError(f'Type {type(self.data)} is not streamable, should be `Generator` or `AsyncGenerator`.')

        if isinstance(self.data, Generator):
            self.data = to_async_generator(self.data)

        async for chunk in self.data:
            if isinstance(chunk, bytes):
                yield chunk
            elif chunk is None:
                yield b''
            else:
                yield json.dumps(chunk)

    async def send(self, send, receive):
        # Send Headers
        await send({'type': 'http.response.start', 'status': self.status_code, 'headers': self.bytes_headers})
        # Send Body as chunks
        asyncio.create_task(self.listen_to_disconnection(receive=receive))
        async for chunk in self.body:
            if self.connection_closed:
                break
            await send({'type': 'http.response.body', 'body': chunk, 'more_body': True})
        else:
            await send({'type': 'http.response.body', 'body': b'', 'more_body': False})


class HTMLResponse(Response):
    """
    Usage Example:
        from panther.response import HTMLResponse

        def my_api():
            html = "<h1>Hello, World!</h1>"
            return HTMLResponse(data=html)
    """

    content_type = 'text/html; charset=utf-8'

    @property
    def body(self) -> bytes:
        if isinstance(self.data, bytes):
            return self.data
        return self.data.encode()


class PlainTextResponse(Response):
    """
    Usage Example:
        from panther.response import PlainTextResponse

        def my_api():
            return PlainTextResponse(data="Hello, World!")
    """

    content_type = 'text/plain; charset=utf-8'

    @property
    def body(self) -> bytes:
        if isinstance(self.data, bytes):
            return self.data
        return self.data.encode()


class TemplateResponse(HTMLResponse):
    """
    Usage Example:
        from panther.response import TemplateResponse

        def my_api():
            context = {"name": "Ali"}
            return TemplateResponse(name="index.html", context=context)
    """

    """
    You may want to declare `TEMPLATES_DIR` in your configs, default is '.'

    Example:
        TEMPLATES_DIR = 'templates/'

    """

    def __init__(
        self,
        source: str | LiteralString | NoneType = None,
        name: str | NoneType = None,
        context: dict | NoneType = None,
        headers: dict | NoneType = None,
        status_code: int = status.HTTP_200_OK,
    ):
        """
        :param source: should be a string
        :param name: name of the template file (should be with its extension, e.g. index.html)
        :param context: should be dict of items
        :param headers: should be dict of headers
        :param status_code: should be int
        """
        if name:
            try:
                template = config.JINJA_ENVIRONMENT.get_template(name=name)
            except jinja2.exceptions.TemplateNotFound:
                loaded_path = ' - '.join(
                    ' - '.join(loader.searchpath)
                    for loader in config.JINJA_ENVIRONMENT.loader.loaders
                    if isinstance(loader, jinja2.loaders.FileSystemLoader)
                )
                error = (
                    f'`{name}` Template Not Found.\n'
                    f'* Make sure `TEMPLATES_DIR` in your configs is correct, Current is {loaded_path}'
                )
                logger.error(error)
                raise APIError
        else:
            template = config.JINJA_ENVIRONMENT.from_string(source=source)
        super().__init__(
            data=template.render(context or {}),
            headers=headers,
            status_code=status_code,
        )


class RedirectResponse(Response):
    """
    Usage Example:
        from panther.response import RedirectResponse

        def my_api():
            return RedirectResponse(url="/new-location")
    """

    def __init__(
        self,
        url: str,
        headers: dict | None = None,
        status_code: int = status.HTTP_307_TEMPORARY_REDIRECT,
        set_cookies: list[Cookie] | None = None,
    ):
        headers = headers or {}
        headers['Location'] = url
        super().__init__(
            headers=headers,
            status_code=status_code,
            set_cookies=set_cookies,
        )



from unittest import IsolatedAsyncioTestCase

from panther import Panther
from panther.app import API
from panther.configs import config
from panther.request import Request
from panther.test import APIClient


@API()
async def flat_multipart_api(request: Request):
    return request.data


@API()
async def single_file_multipart_api(request: Request):
    return {
        'file_name': request.data['file'].file_name,
        'content_type': request.data['file'].content_type,
        'file': request.data['file'].file.decode(),
    }


@API()
async def several_file_multipart_api(request: Request):
    return {
        'file1': {
            'file_name': request.data['file1'].file_name,
            'content_type': request.data['file1'].content_type,
            'file': request.data['file1'].file.decode(),
        },
        'file2': {
            'file_name': request.data['file2'].file_name,
            'content_type': request.data['file2'].content_type,
            'file': request.data['file2'].file.decode(),
        },
    }


@API()
async def complex_multipart_api(request: Request):
    return {
        'name': request.data['name'],
        'age': request.data['age'],
        'file1': {
            'file_name': request.data['file1'].file_name,
            'content_type': request.data['file1'].content_type,
            'file': request.data['file1'].file.decode(),
        },
        'file2': {
            'file_name': request.data['file2'].file_name,
            'content_type': request.data['file2'].content_type,
            'file': request.data['file2'].file.decode(),
        },
    }


@API()
async def multipart_api(request: Request):
    return request.data


urls = {
    'flat_multipart': flat_multipart_api,
    'single_file_multipart': single_file_multipart_api,
    'several_file_multipart': several_file_multipart_api,
    'complex_multipart': complex_multipart_api,
    'multiline_multipart': multipart_api,
}


class TestMultipart(IsolatedAsyncioTestCase):
    CONTENT_TYPE_1 = 'multipart/form-data; boundary=--------------------------201301649688174364392792'
    CONTENT_TYPE_2 = 'multipart/form-data; boundary=----geckoformboundaryc30219e1237602175b34337f41ace019'
    FLAT_PAYLOAD = (
        b'----------------------------201301649688174364392792\r\n'
        b'Content-Disposition: form-data; name="name"\r\n\r\n'
        b'Ali Rn\r\n'
        b'----------------------------201301649688174364392792\r\n'
        b'Content-Disposition: form-data; name="age"\r\n\r\n'
        b'25\r\n'
        b'----------------------------201301649688174364392792--\r\n'
    )
    SINGLE_FILE_PAYLOAD = (
        b'----------------------------201301649688174364392792\r\n'
        b'Content-Disposition: form-data; name="file"; filename="hello_world.txt"\r\n'
        b'Content-Type: text/plain\r\n\r\n'
        b'Hello World\n\r\n'
        b'----------------------------201301649688174364392792--\r\n'
    )
    SEVERAL_FILE_PAYLOAD = (
        b'----------------------------201301649688174364392792\r\n'
        b'Content-Disposition: form-data; name="file1"; filename="hello_world1.txt"\r\n'
        b'Content-Type: text/plain\r\n\r\n'
        b'Hello World1\n\r\n'
        b'----------------------------201301649688174364392792\r\n'
        b'Content-Disposition: form-data; name="file2"; filename="hello_world2.txt"\r\n'
        b'Content-Type: text/plain\r\n\r\n'
        b'Hello World2\n\r\n'
        b'----------------------------201301649688174364392792--\r\n'
    )
    COMPLEX_PAYLOAD = (
        b'----------------------------201301649688174364392792\r\n'
        b'Content-Disposition: form-data; name="name"\r\n\r\n'
        b'Ali Rn\r\n'
        b'----------------------------201301649688174364392792\r\n'
        b'Content-Disposition: form-data; name="file1"; filename="hello_world1.txt"\r\n'
        b'Content-Type: text/plain\r\n\r\n'
        b'Hello World1\n\r\n'
        b'----------------------------201301649688174364392792\r\n'
        b'Content-Disposition: form-data; name="file2"; filename="hello_world2.txt"\r\n'
        b'Content-Type: text/plain\r\n\r\n'
        b'Hello World2\n\r\n'
        b'----------------------------201301649688174364392792\r\n'
        b'Content-Disposition: form-data; name="age"\r\n\r\n'
        b'25\r\n'
        b'----------------------------201301649688174364392792--\r\n'
    )
    MULTI_LINE_PAYLOAD = (
        b'------geckoformboundaryc30219e1237602175b34337f41ace019\r\n'
        b'Content-Disposition: form-data; name="team"\r\n\r\n'
        b'SRE\r\n'
        b'------geckoformboundaryc30219e1237602175b34337f41ace019\r\n'
        b'Content-Disposition: form-data; name="phone"\r\n\r\n'
        b'09033333333\r\n'
        b'------geckoformboundaryc30219e1237602175b34337f41ace019\r\n'
        b'Content-Disposition: form-data; name="message"\r\n\r\n'
        b'My\r\nName\r\nIs\r\nAli\r\n\r\n'
        b'------geckoformboundaryc30219e1237602175b34337f41ace019--\r\n'
    )

    @classmethod
    def setUpClass(cls) -> None:
        app = Panther(__name__, configs=__name__, urls=urls)
        cls.client = APIClient(app=app)

    @classmethod
    def tearDownClass(cls):
        config.refresh()

    async def test_flat_multipart(self):
        res = await self.client.post(
            'flat_multipart',
            content_type=self.CONTENT_TYPE_1,
            payload=self.FLAT_PAYLOAD,
        )
        assert res.status_code == 200
        assert res.data == {'name': 'Ali Rn', 'age': '25'}

    async def test_single_file_multipart(self):
        res = await self.client.post(
            'single_file_multipart',
            content_type=self.CONTENT_TYPE_1,
            payload=self.SINGLE_FILE_PAYLOAD,
        )

        assert res.status_code == 200
        assert res.data == {
            'file_name': 'hello_world.txt',
            'content_type': 'text/plain',
            'file': 'Hello World\n',
        }

    async def test_several_file_multipart(self):
        res = await self.client.post(
            'several_file_multipart',
            content_type=self.CONTENT_TYPE_1,
            payload=self.SEVERAL_FILE_PAYLOAD,
        )

        assert res.status_code == 200
        assert res.data == {
            'file1': {
                'file_name': 'hello_world1.txt',
                'content_type': 'text/plain',
                'file': 'Hello World1\n',
            },
            'file2': {
                'file_name': 'hello_world2.txt',
                'content_type': 'text/plain',
                'file': 'Hello World2\n',
            },
        }

    async def test_complex_multipart(self):
        res = await self.client.post(
            'complex_multipart',
            content_type=self.CONTENT_TYPE_1,
            payload=self.COMPLEX_PAYLOAD,
        )

        assert res.status_code == 200
        assert res.data == {
            'name': 'Ali Rn',
            'age': '25',
            'file1': {
                'file_name': 'hello_world1.txt',
                'content_type': 'text/plain',
                'file': 'Hello World1\n',
            },
            'file2': {
                'file_name': 'hello_world2.txt',
                'content_type': 'text/plain',
                'file': 'Hello World2\n',
            },
        }

    async def test_multiline_multipart(self):
        res = await self.client.post(
            'multiline_multipart',
            content_type=self.CONTENT_TYPE_2,
            payload=self.MULTI_LINE_PAYLOAD,
        )
        assert res.data == {'team': 'SRE', 'phone': '09033333333', 'message': 'My\r\nName\r\nIs\r\nAli\r\n'}



import asyncio
import contextlib
from abc import abstractmethod
from typing import TYPE_CHECKING, Any

from pantherdb import PantherDB

from panther.cli.utils import import_error
from panther.configs import config
from panther.utils import Singleton

try:
    from redis.asyncio import Redis as _Redis
except ImportError:
    # This '_Redis' is not going to be used,
    #   If user really wants to use redis,
    #   we are going to force him to install it in `panther._load_configs.load_redis`
    _Redis = type('_Redis', (), {'__new__': lambda x: x})

if TYPE_CHECKING:
    from pymongo.database import Database


class BaseDatabaseConnection:
    def __init__(self, *args, **kwargs):
        """Initialized in application startup"""
        self.init(*args, **kwargs)

    @abstractmethod
    def init(self, *args, **kwargs):
        pass

    @property
    @abstractmethod
    def session(self):
        pass


class MongoDBConnection(BaseDatabaseConnection):
    def init(
        self,
        host: str = 'localhost',
        port: int = 27017,
        document_class: dict[str, Any] | None = None,
        tz_aware: bool | None = None,
        connect: bool | None = None,
        type_registry=None,  # type: bson.codec_options.TypeRegistry
        database: str | None = None,
        **kwargs: Any,
    ) -> None:
        try:
            from motor.motor_asyncio import AsyncIOMotorClient
        except ModuleNotFoundError as e:
            raise import_error(e, package='motor')

        with contextlib.suppress(ImportError):
            import uvloop

            asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())

        self._client: AsyncIOMotorClient = AsyncIOMotorClient(
            host=host,
            port=port,
            document_class=document_class,
            tz_aware=tz_aware,
            connect=connect,
            type_registry=type_registry,
            **kwargs,
        )
        self._database: Database = self._client.get_database(name=database)

    @property
    def session(self):
        return self._database

    @property
    def client(self):
        return self._client


class PantherDBConnection(BaseDatabaseConnection):
    def init(self, path: str | None = None, encryption: bool = False):
        params = {'db_name': path, 'return_dict': True, 'return_cursor': True}
        if encryption:
            try:
                import cryptography
            except ImportError as e:
                raise import_error(e, package='cryptography')
            params['secret_key'] = config.SECRET_KEY.encode()

        self._connection: PantherDB = PantherDB(**params)

    @property
    def session(self):
        return self._connection

    @property
    def client(self):
        return self._connection


class DatabaseConnection(Singleton):
    @property
    def session(self):
        return config.DATABASE.session

    @property
    def is_defined(self):
        return bool(config.DATABASE)

    @property
    def client(self):
        return config.DATABASE.client


class RedisConnection(Singleton, _Redis):
    is_connected: bool = False

    def __init__(
        self,
        init: bool = False,
        host: str = 'localhost',
        port: int = 6379,
        db: int = 0,
        **kwargs,
    ):
        if init:
            self.host = host
            self.port = port
            self.db = db
            self.kwargs = kwargs

            super().__init__(host=host, port=port, db=db, **kwargs)
            self.is_connected = True
            self.sync_ping()

    def sync_ping(self):
        from redis import Redis

        Redis(host=self.host, port=self.port, socket_timeout=3, **self.kwargs).ping()

    async def execute_command(self, *args, **options):
        if self.is_connected:
            return await super().execute_command(*args, **options)
        msg = '`REDIS` is not found in `configs`'
        raise ValueError(msg)

    def create_connection_for_websocket(self) -> _Redis:
        if not hasattr(self, 'websocket_connection'):
            self.websocket_connection = _Redis(host=self.host, port=self.port, db=0, **self.kwargs)
        return self.websocket_connection


db: DatabaseConnection = DatabaseConnection()
redis: RedisConnection = RedisConnection()



from abc import abstractmethod

from panther.request import Request
from panther.websocket import Websocket


class BasePermission:
    @abstractmethod
    async def __call__(self, request: Request | Websocket) -> bool:
        return True


class IsAuthenticated(BasePermission):
    async def __call__(self, request: Request) -> bool:
        return bool(request.user)


class IsAuthenticatedOrReadonly(BasePermission):
    async def __call__(self, request: Request) -> bool:
        return request.user or request.method == 'GET'



import logging
import sys

from panther import version as panther_version
from panther.cli.create_project_command import create
from panther.cli.create_user_command import create_user
from panther.cli.monitor_command import monitor
from panther.cli.run_command import run
from panther.cli.utils import cli_error, print_help_message

logger = logging.getLogger('panther')


def shell(args) -> None:
    if len(args) == 0:
        return cli_error(
            'Not Enough Arguments, Give me a file path that contains `Panther()` app.\n'
            '       * Make sure to run `panther shell` in the same directory as that file!\n'
            '       * Example: `panther shell main.py`',
        )
    elif len(args) != 1:
        return cli_error('Too Many Arguments.')

    package = args[0].removesuffix('.py')
    try:
        from IPython import start_ipython

        start_ipython(('--gui', 'asyncio', '-c', f'"import {package}"', '-i'))
    except ImportError:
        logger.error('Make sure `ipython` is installed -> Hint: `pip install ipython`')


def version() -> None:
    print(panther_version())


def start() -> None:
    args = sys.argv[2:]
    match len(sys.argv) > 1 and sys.argv[1]:
        case 'h' | 'help' | '-h' | '--help':
            print_help_message()
        case 'create':
            create(args)
        case 'createuser':
            create_user(args)
        case 'run':
            run(args)
        case 'shell':
            shell(args)
        case 'monitor':
            monitor()
        case 'version' | '--version':
            version()
        case _:
            cli_error('Invalid Arguments.')



from panther.middlewares import HTTPMiddleware
from panther.request import Request
from panther.response import RedirectResponse


class RedirectToSlashMiddleware(HTTPMiddleware):
    async def __call__(self, request: Request):
        if not request.path.endswith('/'):
            return RedirectResponse(request.path + '/')
        return await self.dispatch(request=request)
